# Check for required modules:
import sys
import os
from .module_check import module_check
module_check()

import tkinter as tk
from tkinter import messagebox
from tkinter import filedialog
from tkinter import font

from .tao_widget import *
from .tao_set import tao_set
from pytao.util.parameters import str_to_tao_param, tao_startup_param_dict
from pytao.util.parameters import tao_parameter_dict
from .tao_console import tao_console
from .tao_plot_dict import *
from .tao_interface import tao_interface

from .tao_data_windows import *
from .tao_lat_windows import *
from .tao_misc_windows import *
from .tao_plot_windows import *
from .tao_var_windows import *
from .tao_beam_windows import *

#---------------------------------------------------------------
# Root window

class tao_root_window(tk.Tk):

    def __init__(self, *args, **kwargs):
        tk.Tk.__init__(self, className='Tao')

        self.title("Tao")
        self.protocol("WM_DELETE_WINDOW", self.quit_cmd)
        self.tk.call('tk', 'scaling', 1.0)
        #default_font = font.nametofont("TkDefaultFont")
        #default_font.configure(size=14)
        #self.option_add("*Font", default_font)

        self.GUI_DIR = (os.environ['ACC_LOCAL_ROOT'] if
                'ACC_LOCAL_ROOT' in os.environ.keys() else os.environ['ACC_ROOT_DIR'])
        self.GUI_DIR += '/tao/python/pytao/gui'
        if sys.platform == "linux": # setting icon only works on linux for now
            self.icon = '@' + self.GUI_DIR + "/assets/tao-logo.xbm"
            self.iconbitmap(self.icon)

        # Window lists (accessible for refreshing)
        self.refresh_windows = {}
        self.tao_id_list = ['plot', 'data', 'var', 'ele', 'lat']
        for w in self.tao_id_list:
            self.refresh_windows[w] = []

        # Key bindings

        self.bind_all("<Control-q>", self.quit_cmd)

        # Init GUI

        self.old_init_args = ""
        init_frame = tk.Frame(self)
        init_frame.pack()
        self.tao_load(init_frame)

    #-----------------------------------------------------------------------------------
    # Tao startup

    def start_main(self):
        self.unbind("<Return>")
        self.lift()
        self.focus_force()
        #self.menubar.entryconfig("File", state="normal")
        #self.menubar.entryconfig("Window", state="normal")
        self.menubar_init()
        self.main_frame = tk.Frame(self, width = 20, height = 30)
        self.main_frame.pack()
        #Key bindings
        self.bind_all('<Control-n>', self.new_data_event)
        self.bind_all('<Alt-n>', self.new_var_event)
        self.bind_all('<Alt-p>', self.new_template_event)
        self.bind_all('<Control-g>', self.global_vars_event)
        self.bind_all('<Alt-g>', self.bmad_com_event)
        self.bind_all('<Control-d>', self.view_data_event)
        self.bind_all('<Control-e>', self.view_ele_event)
        self.bind_all('<Control-l>', self.view_lattice_event)
        self.bind_all('<Control-t>', self.plot_template_event)
        self.bind_all('<Control-r>', self.plot_region_event)
        self.bind_all('<Alt-v>', self.view_vars_event)
        self.bind_all('<Alt-q>', self.reinit_event)
        self.bind_all('<Control-b>', self.beam_init_event)
        # Clear out python lat_calc_done
        self.pipe.cmd_in('python lat_calc_done')

        # Call
        tk.Label(self.main_frame, text="Call command file:").grid(row=0, column=0)
        self.call_file = tk_tao_parameter(str_to_tao_param("call_file;FILE;T;"), self.main_frame, self.pipe)
        self.call_file.tk_wid.grid(row=0, column=1, sticky='EW')
        tk.Button(self.main_frame, text="Run", command=self.tao_call).grid(row=0, column=2, sticky='EW')
        # Arguments
        tk.Label(self.main_frame, text="Arguments:").grid(row=1, column=0)
        self.cf_args = tk_tao_parameter(str_to_tao_param("cf_args;STR;T;"), self.main_frame, self.pipe)
        self.cf_args.tk_wid.configure(width=30)
        self.cf_args.tk_wid.bind("<Return>", self.tao_call)
        self.cf_args.tk_wid.grid(row=1, column=1, columnspan=2, sticky='EW')

        # Optimization
        tk.Label(self.main_frame, text="Optimization:").grid(row=2, column=0)
        tk.Button(self.main_frame, text="Setup").grid(row=2, column=1, sticky='EW')
        tk.Button(self.main_frame, text="Run").grid(row=2, column=2, sticky='EW')

        # Debug
        tk.Button(self.main_frame, text="Debug", command=self.debug_cmd).grid(
                row=3, column=0, columnspan=3, sticky='EW')

        # Command line
        self.cmd_frame = tk.Frame(self)
        self.cmd_frame.pack(side="bottom", fill='both', expand=1)
        self.history = [] #holds the history
        self.history.append([]) #Tao and shell history
        self.history_pos = 0 #Used for scrolling in history on command line
        self.history.append([]) #Call history
        tk.Button(self.cmd_frame, text="View History...",
                command=self.view_history_cmd).grid(row=0, column=1, sticky='EW')
        tk.Button(self.cmd_frame, text="Show/Hide Console",
                command=self.sh_console).grid(row=0, column=0, sticky='EW')
        self.console = tao_console(self.cmd_frame, self, self.pipe)
        self.cmd_frame.columnconfigure(0, weight=1)
        self.cmd_frame.columnconfigure(1, weight=1)
        self.console.grid(row=1, column=0, columnspan=2, sticky = 'NSEW')
        self.cmd_frame.rowconfigure(1, weight=1)
        self.console_packed = True

    #-----------------------------------------------------------------------------------

    def sh_console(self, event=None):
        '''
        Packs or unpacks self.console as appropriate
        '''
        if self.console_packed:
            self.console.grid_forget()
        else:
            self.console.grid(row=1, column=0, columnspan=2, sticky = 'NSEW')
        self.console_packed = not self.console_packed

    #-----------------------------------------------------------------------------------

    def tao_call(self, event=None):
        '''
        Runs the command file in self.call_file, appends it to the history,
        and clears self.call_file
        '''
        if self.call_file.tk_var.get() != "Browse...":
            self.pipe.cmd("call " + self.call_file.tk_var.get() + ' '
                    + self.cf_args.tk_var.get())
            self.history[1].append(self.call_file.tk_var.get())
            self.call_file.tk_var.set("Browse...")
        #Try to refresh history window
        try:
            self.history_window.refresh()
        except:
            pass
        # Check the place buffer and place plots
        self.default_plots(include_init=False)
        # Update pgplot settings
        if self.plot_mode == "pgplot/plplot":
            self.placed.pgplot_update()

    #-----------------------------------------------------------------------------------

    def menubar_init(self):
        self.menubar = tk.Menu(self)

        file_menu = tk.Menu(self.menubar, tearoff=0)
        file_menu.add_command(label = 'Write Data Namelist...', command = self.write_data_cmd)
        file_menu.add_command(label = 'Write Var Namelist...', command = self.write_var_cmd)
        file_menu.add_separator()
        file_menu.add_command(label = 'Reinit...',
                command = self.reinit_cmd,accelerator = 'Alt+Q')
        file_menu.add_command(label = 'Quit', command = self.quit_cmd,
                accelerator = 'Ctrl+Q')
        self.menubar.add_cascade(label = 'File', menu = file_menu)

        create_menu = tk.Menu(self.menubar, tearoff=0)
        create_menu.add_command(label = "New Data...",
                command = self.new_data_cmd, accelerator = 'Ctrl+N')
        create_menu.add_command(label = "New Variables...",
                command = self.new_var_cmd, accelerator = 'Alt+N')
        create_menu.add_command(label = "New Plot Template...",
                command = self.new_template_cmd, accelerator = 'Alt+P')
        self.menubar.add_cascade(label = 'Create', menu = create_menu)

        view_menu = tk.Menu(self.menubar, tearoff=0)
        view_menu.add_command(label = 'Data...',
                command = self.view_data_cmd, accelerator = 'Ctrl+D')
        view_menu.add_command(label = 'Variables...',
                command = self.view_vars_cmd, accelerator = 'Alt+V')
        view_menu.add_command(label = 'Lattice Elements...',
                command = self.view_ele_cmd, accelerator = 'Ctrl+E')
        view_menu.add_command(label = 'Lattice...',
                command = self.view_lattice_cmd, accelerator = 'Ctrl+L')
        view_menu.add_command(label = 'Lat_layout Shapes...',
                command = self.lat_layout_shape_cmd)
        view_menu.add_command(label = 'Floor_plan Shapes...',
                command = self.floor_plan_shape_cmd)
        view_menu.add_command(label = 'Global Variables...',
                command = self.set_global_vars_cmd, accelerator = 'Ctrl+G')
        view_menu.add_command(label = 'Bmad Parameters...',
                command = self.set_bmad_com_cmd, accelerator = 'Alt+G')
        self.menubar.add_cascade(label = 'View', menu = view_menu)

        plot_menu = tk.Menu(self.menubar, tearoff=0)
        plot_menu.add_command(label = 'New Plot...',
                command = self.plot_template_cmd, accelerator = 'Ctrl+T')
        plot_menu.add_command(label = 'Edit Plot...',
                command = self.plot_region_cmd, accelerator = 'Ctrl+R')
        plot_menu.add_command(label = 'Building Wall Settings...',
                command = self.building_wall_cmd)
        self.menubar.add_cascade(label = 'Plot', menu = plot_menu)

        beam_menu = tk.Menu(self.menubar, tearoff=0)
        beam_menu.add_command(label = 'Beam settings...',
                command = self.beam_init_cmd, accelerator = 'Ctrl+B')
        self.menubar.add_cascade(label = 'Beam', menu = beam_menu)

        #window_menu = tk.Menu(self.menubar)
        #window_menu.add_command(label = 'Optimizer...',
        #        command = self.optimizer_cmd)
        #window_menu.add_command(label = 'Wave...', command = self.wave_cmd)
        #self.menubar.add_cascade(label = 'Window', menu = window_menu)

        self.config(menu=self.menubar)

    #-----------------------------------------------------------------------------------

    def default_plots(self, include_init=True):
        '''
        If self.plot_mode == "matplotlib", opens all the
        plot templates listed in plot.gui.init in
        separate matplotlib windows
        Also opens matplotlib windows for any plots placed by the tao init file
        Will only place plots from the place buffer if include_init is False
        '''
        if include_init:
            # Open plot.gui.init
            try:
                plot_file = open('plot.gui.init')
                plot_list = plot_file.read()
                plot_list = plot_list.splitlines()
                plot_file.close()
            except:
                plot_list = []
        else:
            plot_list = []
        # Read the place buffer
        init_plots = self.pipe.cmd_in('python place_buffer')
        init_plots = init_plots.splitlines()
        init_regions = [None]*len(init_plots)
        for i in range(len(init_plots)):
            init_regions[i] = init_plots[i].split(';')[0]
            init_plots[i] = init_plots[i].split(';')[1]
        plot_list = init_plots + plot_list
        plot_regions = init_regions + [None]*len(init_plots)

        # Get list of plot templates to check input against
        plot_templates = self.pipe.cmd_in("python plot_list t")
        plot_templates = plot_templates.splitlines()
        for i in range(len(plot_templates)):
            plot_templates[i] = plot_templates[i].split(';')[1]

        # Place the plots
        # Validate entries in plot.gui.init
        for i in range(len(plot_list)):
            plot = plot_list[i]
            if plot in plot_templates:
                if self.plot_mode == "matplotlib":
                    # Make a window for the plot
                    win = tao_plot_window(self, plot, self.pipe, region=plot_regions[i])
                elif self.plot_mode == "pgplot/plplot":
                    # self.placed (tao_plot_dict) will keep track of how many
                    # rows/columns of the pgplot window are available
                    self.placed.place_template(plot, plot_regions[i])

        # Determine pgplot layout from currently placed plots
        if self.plot_mode == "pgplot/plplot":
            self.placed.pgplot_update()


    #-----------------------------------------------------------------------------------

    def tao_load(self, init_frame):
        '''
        Handles the startup of tao, including parsing gui.init and commandline
        arguments.
        '''

        def toggle_init(*args):
            '''
            Enables/disables the init file box if noinit is selected
            '''
            ix_noinit = tk_list.index('noinit')
            ix_init_file = tk_list.index('init_file')
            if tk_list[ix_noinit].tk_var.get():
                tk_list[ix_init_file].tk_wid.configure(state='disabled')
            else:
                tk_list[ix_init_file].tk_wid.configure(state='normal')
        # End: def toggle_init

        def swap_box(event=None):
            if interface_to_tao.get() == "pexpect":
                ctype_label.grid_forget()
                tao_lib.tk_wid.grid_forget()
                pexp_label.grid(row=k+1, sticky="E")
                tao_exe.tk_wid.grid(row=k+1, column=1, sticky="W")
            elif interface_to_tao.get() == "ctypes":
                pexp_label.grid_forget()
                tao_exe.tk_wid.grid_forget()
                ctype_label.grid(row=k+1, sticky="E")
                tao_lib.tk_wid.grid(row=k+1, column=1, sticky="W")
        # End: def swap_box

        def param_load(event=None):
            # set the font
            try:
                new_size = int(font_var.get())
                new_font = font.nametofont("TkDefaultFont")
                new_font.configure(size=new_size)
                self.option_add("*Font", new_font)
                self.font_size = new_size
            except:
                self.font_size = None
            # set self.lw
            self.lw = lw_var.get()
            self.plot_mode = plot_mode.get()
            # With noinit, make sure lattice file is chosen
            ix_noinit = tk_list.index('noinit')
            ix_lat_file = tk_list.index('lattice_file')
            if tk_list[ix_noinit].tk_var.get() and tk_list[ix_lat_file].tk_var.get() in ["", "Browse..."]:
                messagebox.showwarning("Warning", "A lattice file is required when starting Tao without an init file.")
                return
            init_args = ""
            for tk_param in tk_list:
                if tk_param.param.type in ['STR','ENUM']:
                    tk_param.param.value = tk_param.tk_var.get()
                    if tk_param.param.value != "":
                        init_args = (init_args + "-" + tk_param.param.name + " "
                                + tk_param.param.value + " ")
                    elif tk_param.param.name in self.old_init_args:
                        init_args += '--' + tk_param.param.name + ' '
                elif tk_param.param.type == 'FILE':
                    tk_param.param.value = tk_param.tk_var.get()
                    if tk_param.param.value == "Browse...":
                        tk_param.param.value = ""
                    if tk_param.param.value != "":
                        init_args = (init_args + "-" + tk_param.param.name + " "
                                + tk_param.param.value + " ")
                    elif tk_param.param.name in self.old_init_args:
                        init_args += '--' + tk_param.param.name + ' '
                elif tk_param.param.type == 'LOGIC':
                    tk_param.param.value = tk_param.tk_var.get()
                    if tk_param.param.value:
                        init_args = init_args + "-" + tk_param.param.name + " "
                    elif tk_param.param.name in self.old_init_args:
                        init_args += '--' + tk_param.param.name + ' '
            #
            if plot_mode.get() != "pgplot/plplot":
                init_args = init_args + "-noplot -external_plotting"
            # Run Tao, clear the init_frame, and draw the main frame
            if interface_to_tao.get() == "pexpect":
                mode = "pexpect"
                if tao_exe.tk_var.get() == "Browse...":
                    tao_exe.tk_var.set("")
                self.pipe = tao_interface(mode, init_args, tao_exe.tk_var.get())
            elif interface_to_tao.get() == "ctypes":
                mode = "ctypes"
                if tao_lib.tk_var.get() == "Browse...":
                    tao_lib.tk_var.set("")
                #Ctypes needs a more sophisticated reinit using reinit tao
                #Check if Tao is open
                if 'pipe' in self.__dict__ and self.pipe.mode == 'ctypes':
                    self.pipe.cmd_in('reinit tao ' + init_args)
                else:
                    self.pipe = tao_interface(mode, init_args, so_lib=tao_lib.tk_var.get())
            self.old_init_args = init_args

            init_frame.destroy()
            self.start_main()
            # Clear the plotting lists
            self.placed = tao_plot_dict(self.plot_mode, self.pipe)
            if plot_mode.get() == "matplotlib":
                # place the lattice layout in layout by default
                self.placed.place_template('lat_layout', 'layout')
            self.default_plots()
        # End: def param_load

        def update_font(event=None):
            try:
                new_size = int(font_var.get())
                new_font = font.nametofont("TkDefaultFont")
                new_font.configure(size=new_size)
                self.option_add("*Font", new_font)
            except:
                pass
            return 'break' # Do not load into Tao yet

        # End of local defs
        #-----------------------

        tk_list = [] #Items: tk_tao_parameter() (see tao_widget.py)
        init_frame.grid_columnconfigure(0, weight=1, pad=10)
        init_frame.grid_columnconfigure(1, weight=1)
        warning_messages = []
        gui_init_file = None

        # Parse command line arguments
        clargs = {} # keys: command line switches, values: switch value
        looking_for = 'switch'
        for i in range(len(sys.argv)):
            if looking_for == 'switch':
                arg = sys.argv[i]
                if arg[0] == '-': # -switch
                    arg = arg[1:]

                    normal = True
                    if arg[0] == '-':   # "--switch" negates the effect of "-switch"
                      normal = False
                      arg = arg[1:]

                    # Determine if arg is a valid switch name
                    matches = [] # list of startup params that arg might refer to
                    # Special handling for gui_init
                    if arg == 'gui_init':
                        gui_init_file = ""
                        looking_for = 'file'
                        continue
                    # Generic case
                    for param in tao_startup_param_dict.keys():
                        if param.find(arg) == 0:
                            matches.append(param)
                    if len(matches) == 1: #exactly one match -> good switch
                        arg = matches[0]
                        if (tao_startup_param_dict[matches[0]].type == 'FILE') & (i != len(sys.argv)-1):
                            clargs[arg] = "" #add arg to clargs
                            if normal: looking_for = 'file' # file switches need files
                        elif tao_startup_param_dict[matches[0]].type == 'LOGIC':
                            if normal:
                              clargs[arg] = "T" #add arg to clargs
                            else:
                              clargs[arg] = "F" #add arg to clargs
            elif looking_for == 'file':
                arg_file = sys.argv[i]
                if arg_file.find('-') == 0: #not a file
                    if arg == 'gui_init':
                        pass
                    else:
                        # Remove previous argument from list
                        clargs.pop(arg)
                else:
                    if arg == 'gui_init':
                        gui_init_file = arg_file
                    else:
                        clargs[arg] = arg_file
                looking_for = 'switch'

        #Look for and read gui.init
        #each line should have the form
        #parameter:value
        #E.g. init_file:tao.init
        #E.g. rf_on:T
        if gui_init_file == None:
            # Try to use gui.init if init file not specified
            gui_init_file = os.path.expanduser('gui.init')
        else:
            gui_init_file = os.path.expanduser(gui_init_file)
            gui_init_file = os.path.expandvars(gui_init_file)

        try:
            init_file = open(gui_init_file)
            init_list = init_file.read()
            init_list = init_list.splitlines()
            init_file.close()
        except:
            init_list = []

        for key, value in clargs.items(): # add clargs to init_list
            init_list.append(key + ':' + value)

        init_dict = {}
        for entry0 in init_list:
            entry = entry0.strip()
            entry = entry.split('#', 1)[0]   # Remove comment
            if entry == '': continue
            #Check for proper formatting and not a comment
            if entry.find(':') != -1:
                name = entry[:entry.find(':')]
                name = name.strip()
                value = entry[entry.find(':')+1:]
                value = value.strip()
            else:
                messagebox.showwarning('Error!', 'Bad line in gui.init file:\n' + entry0)
                continue

            c1 = (name in ["tao_executable", "tao_shared_library"])
            try:
                c2 = (tao_startup_param_dict[name].type == 'FILE')
            except KeyError:
                c2 = False
            if c1 | c2:
                #Check that a good filename has been given
                try:
                    filename = value
                    #Expand environment variables and ~
                    filename = os.path.expanduser(filename)
                    filename = os.path.expandvars(filename)
                    f = open(filename)
                    f.close()
                    init_dict[name] = filename
                except FileNotFoundError:
                    #Don't put bad files in init_dict
                    #messagebox.showwarning(name, "File not found: " + filename)
                    warning_messages.append([name, filename])
            else:
                init_dict[name] = value

        k = 0 #row number counter
        for param, tao_param in tao_startup_param_dict.items():
            if param == 'noplot': continue   # The setting of noplot affects the "Plot Mode"
            tk_list.append(tk_tao_parameter(tao_param,init_frame))
            #Possibly set value from init file
            if tk_list[k].param.name in init_dict:
                if tk_list[k].param.type == 'FILE':
                    tk_list[k].tk_var.set(init_dict[tk_list[k].param.name])
                elif tk_list[k].param.type == 'LOGIC':
                    state = (init_dict[tk_list[k].param.name] == 'T') | (init_dict[tk_list[k].param.name] == 'True')
                    tk_list[k].tk_var.set(state)

            if param == 'noinit': #gets a special label
                tk.Label(init_frame, text="Don't use an init file (noinit)").grid(row=k,sticky="E")
                tk_list[k].tk_var.trace('w', toggle_init)
            else:
                tk.Label(init_frame, text=param).grid(row=k,sticky="E")

            tk_list[k].tk_wid.grid(row=k, column=1, sticky="W")
            k = k+1

        # Add separator
        tk.ttk.Separator(init_frame, orient="horizontal").grid(row=k, columnspan=2, sticky='ew')
        k = k+1

        # Choosing whether to use pexpect or ctypes must be handled separately
        tk.Label(init_frame, text="Interface to Tao:").grid(row=k, sticky="E")

        interface_to_tao = tk.StringVar()
        interface_to_tao.set("ctypes")
        if "interface_to_tao" in init_dict:
            if init_dict["interface_to_tao"] in ["pexpect", "ctypes"]:
                interface_to_tao.set(init_dict["interface_to_tao"])
        tk.OptionMenu(init_frame, interface_to_tao, "pexpect", "ctypes", command=swap_box).grid(row=k, column=1, sticky="W")
        pexp_label = tk.Label(init_frame, text="Tao Executable")
        ctype_label = tk.Label(init_frame, text="Tao Shared Library")
        tao_exe = tk_tao_parameter(str_to_tao_param("tao_executable;FILE;T;"), init_frame)
        if "tao_executable" in init_dict:
            tao_exe.tk_var.set(init_dict["tao_executable"])
        elif 'ACC_LOCAL_ROOT' in os.environ.keys():
            tao_exe.tk_var.set(os.environ['ACC_LOCAL_ROOT']+'/production/bin/tao')
        elif 'ACC_EXE' in os.environ.keys():
            tao_exe.tk_var.set(os.environ['ACC_EXE']+'/tao')
        tao_lib = tk_tao_parameter(str_to_tao_param("tao_shared_library;FILE;T;"), init_frame)
        if "tao_shared_library" in init_dict:
            tao_lib.tk_var.set(init_dict["tao_shared_library"])
        elif 'ACC_LOCAL_ROOT' in os.environ.keys() and \
                               os.path.isfile(os.environ['ACC_LOCAL_ROOT']+'/production/lib/libtao.so'):
            tao_lib.tk_var.set(os.environ['ACC_LOCAL_ROOT']+'/production/lib/libtao.so')
        elif 'ACC_ROOT_DIR' in os.environ.keys() and \
                                  os.path.isfile(os.environ['ACC_ROOT_DIR']+'/production/lib/libtao.so'):
            tao_lib.tk_var.set(os.environ['ACC_ROOT_DIR']+'/production/lib/libtao.so')
        swap_box()

        #Choosing plot mode must also be handled separately
        tk.Label(init_frame, text="Plot Mode").grid(row=k+2, sticky='E')
        plot_mode = tk.StringVar()
        plot_mode.set("matplotlib")
        plot_options = ["matplotlib", "pgplot/plplot", "none"]
        plot_chooser = tk.OptionMenu(init_frame, plot_mode, *plot_options)
        plot_chooser.grid(row=k+2, column=1, sticky='W')
        # Set plot_mode from init_dict if specified
        if "plot_mode" in init_dict:
            if init_dict["plot_mode"] == "pgplot":
                plot_mode.set("pgplot/plplot")
            elif init_dict["plot_mode"] in plot_options:
                plot_mode.set(init_dict["plot_mode"])
            else:
                messagebox.showwarning('Error', 'Bad setting of "plot_mode" in gui.init: ' + init_dict["plot_mode"])

        if 'noplot' in init_dict and init_dict['noplot'] == 'T': plot_mode.set('none')

        # Font size
        tk.Label(init_frame, text="Font size").grid(row=k+3, sticky='E')
        font_var = tk.StringVar()
        if 'font_size' in init_dict.keys():
            font_var.set(str(init_dict['font_size']))
        else:
            font_var.set('14')
        font_box = tk.Entry(init_frame, textvariable=font_var)

        update_font() # Set the font for the first time
        font_box.bind('<Return>', update_font)
        font_box.grid(row=k+3, column=1, sticky='W')

        # Light-weight switch
        lw_var = tk.BooleanVar()
        if 'lw' in init_dict:
            lw_var.set((init_dict['lw']=='T') | (init_dict['lw'] == 'True'))
        else:
            lw_var.set(True)
        # Start button
        load_b = tk.Button(init_frame, text="Start Tao", command=param_load)
        load_b.grid(row=k+5, columnspan=2)
        self.bind("<Return>", param_load)

        # Show warning messages about bad filenames from startup
        for warning in warning_messages:
            messagebox.showwarning(warning[0], "File not found: " + warning[1])

        #Start Tao immediately if start_tao is set true
        c1 = "start_tao" in init_dict
        c2 = len(warning_messages) == 0
        c3 = 'pipe' not in self.__dict__
        c4 = ('start_tao' in init_dict.keys()) and (init_dict["start_tao"] in ['T', 'True'])
        if c1 & c2 & c3 & c4: param_load()

        if ('do_mainloop' in init_dict.keys()) and (init_dict["do_mainloop"] in ['F', 'False']):
          self.do_mainloop = False
        else:
          self.do_mainloop = True

    #-----------------------------------------------------------------------------------
    # Menu bar callbacks

    def new_data_cmd(self):
        win = tao_new_data_window(self, self.pipe)

    def new_var_cmd(self):
        win = tao_new_var_window(self, self.pipe)

    def new_template_cmd(self):
        win = tao_new_plot_template_window(self, self.pipe)

    def write_data_cmd(self):
        '''
        Writes a fortran namelist file for all d2_arrays
        '''
        out_file = filedialog.asksaveasfilename(title="Save data namelist as...")
        if out_file not in [(), ""]:
            self.pipe.cmd_in('write namelist -data ' + out_file)

    def write_var_cmd(self):
        '''
        Writes a fortran namelist file for all v1_arrays
        '''
        out_file = filedialog.asksaveasfilename(title="Save var namelist as...")
        if out_file not in [(), ""]:
            self.pipe.cmd_in('write namelist -variable ' + out_file)

    def reinit_cmd(self):
        '''
        Quit Tao, destroy the main frame, and respawn the init frame
        '''
        result = messagebox.askquestion("Reinit",
                "This will close all windows and restart Tao.  Are you sure?")
        if result != 'yes':
            return

        for child in self.winfo_children():
            child.destroy()

        if self.pipe.mode == 'pexpect':
            # Hack to prevent pexpect from hanging:
            self.pipe.cmd_in(" spawn echo Tao\\>; quit", no_warn=True)
        init_frame = tk.Frame(self)
        init_frame.pack()
        self.tao_load(init_frame)

    def quit_cmd(self, event = ''):
        result = messagebox.askquestion("Quit", "Are You Sure?", icon='warning')
        if result == 'yes':
            for child in self.winfo_children():
                # better safe than sorry
                try:
                    child.destroy()
                except:
                    pass
            sys.exit(0)
        else:
            return

    def optimizer_cmd(self):
        print ('Optimizer called')
        win = tk.Toplevel(self)
        win.title('Optimizer')

    def beam_init_cmd(self):
        win = tao_beam_init_window(self, self.pipe)

    def plot_template_cmd(self):
        win = tao_place_plot_window(self, self.pipe)

    def plot_region_cmd(self):
        #win = tao_plot_tr_window(self, self.pipe, "R")
        win = tao_new_plot_template_window(self, self.pipe, None, 'T')

    def building_wall_cmd(self):
        win = tao_building_wall_window(self, self.pipe)

    def wave_cmd(self):
        print ('Wave called')

    def set_global_vars_cmd(self):
        win = tao_global_vars_window(self)

    def set_bmad_com_cmd(self):
        win = tao_bmad_com_window(self)

    def view_vars_cmd(self):
        win = tao_var_general_window(self, self.pipe)

    def view_data_cmd(self):
        win = tao_d2_data_window(self, self.pipe)

    def view_ele_cmd(self):
        win = tao_ele_window(self, self.pipe)

    def view_lattice_cmd(self):
        win = tao_lattice_window(self, self.pipe)

    def lat_layout_shape_cmd(self):
        win = tao_ele_shape_window(self, self.pipe, "lat_layout")

    def floor_plan_shape_cmd(self):
        win = tao_ele_shape_window(self, self.pipe, "floor_plan")

    # Other callbacks

    def new_data_event(self, event):
        self.new_data_cmd()

    def new_var_event(self, event):
        self.new_var_cmd()

    def new_template_event(self, event):
        self.new_template_cmd()

    def global_vars_event(self, event):
        self.set_global_vars_cmd()

    def bmad_com_event(self, event):
        self.set_bmad_com_cmd()

    def view_data_event(self, event):
        self.view_data_cmd()

    def view_ele_event(self, event):
        self.view_ele_cmd()

    def view_lattice_event(self, event):
        self.view_lattice_cmd()

    def plot_template_event(self, event):
        self.plot_template_cmd()

    def plot_region_event(self, event):
        self.plot_region_cmd()

    def view_vars_event(self, event):
        self.view_vars_cmd()

    def beam_init_event(self, event):
        self.beam_init_cmd()

    def reinit_event(self, event):
        self.reinit_cmd()

    def view_history_cmd(self):
        #Just focus the existing history window, if it exists
        try:
            self.history_window.lift()
            #self.history_window.force_focus()
        except:
            self.history_window = tao_history_window(self)

    def debug_cmd(self):
        win = tk.Toplevel(self)
        win.title('Debug')

        tk.Label(win, text='Commands to print to stdout:').grid(row=0, column=0, sticky='EW')
        var = tk.StringVar()
        opts = ['None', 'Omit querries (show, python var_general, etc)', 'All']
        i=1
        for o in opts:
            tk.Radiobutton(win, text=o, variable=var, value=str(i)).grid(row=i, column=0, sticky='EW')
            i += 1
        def set_debug():
            '''Command for the button'''
            if var.get() == '1':
                print('1')
                self.pipe.debug = False
                self.pipe.debug_patterns = []
            elif var.get() == '2':
                print(2)
                self.pipe.debug = True
                self.pipe.debug_patterns = ['set', 'place',
                        'python data_d2_create', 'python datum_create',
                        'python var_create', 'python var_v1_create',
                        'python plot_manage_plot',
                        'python plot_manage_curve',
                        'python plot_manage_graph',
                        'python lat_calc_done', 'python place_buffer'
                        'write', 'call']
            elif var.get() == '3':
                print(3)
                self.pipe.debug = True
                self.pipe.debug_patterns = 'All'
        tk.Button(win, text='Apply', command=set_debug).grid(row=i, column=0, sticky='EW')

#-----------------------------------------------------

if __name__ == "__main__":
    root = tao_root_window(sys.argv)
    root.mainloop()



import os
import ctypes
import numpy as np
from pytao import tao_ctypes
from pytao.tao_ctypes import extra_commands
from pytao.tao_ctypes.util import error_in_lines
from pytao.util.parameters import tao_parameter_dict
from .tools import full_path
import tempfile
import shutil
import types


import logging
logger = logging.getLogger(__name__)


#--------------------------------------

class Tao:
    """
    Class to run and interact with Tao. Requires libtao shared object.

    Setup:

    import os
    import sys
    TAO_PYTHON_DIR=os.environ['ACC_ROOT_DIR'] + '/tao/python'
    sys.path.insert(0, TAO_PYTHON_DIR)

    import tao_ctypes
    tao = tao_ctypes.Tao()
    tao.init("command line args here...")
    """

    #---------------------------------------------

    def __init__(self, init='', so_lib=''):
        # TL/DR; Leave this import out of the global scope.
        #
        # Make it lazy import to avoid cyclical dependency.
        # at __init__.py there is an import for Tao which
        # would cause interface_commands to be imported always
        # once we import pytao.
        # If by any chance the interface_commands.py is broken and
        # we try to autogenerate it will complain about the broken
        # interface_commands file.
        
        from pytao import interface_commands 
        
        # Library needs to be set.
        self.so_lib_file = None
        if so_lib == '':
            # Search
            ACC_ROOT_DIR = os.getenv('ACC_ROOT_DIR', '')
            if ACC_ROOT_DIR:
                BASE_DIR = os.path.join(ACC_ROOT_DIR, 'production', 'lib')
                self.so_lib_file = find_libtao(BASE_DIR)
        else:
            self.so_lib_file = so_lib

        if self.so_lib_file:
            self.so_lib = ctypes.CDLL(self.so_lib_file)
        else:
            lib, lib_file = auto_discovery_libtao()

            if lib:
                self.so_lib = lib
                self.so_lib_file = lib_file
            else:
                raise ValueError(f'Shared object libtao library not found.')

        self.so_lib.tao_c_out_io_buffer_get_line.restype = ctypes.c_char_p
        self.so_lib.tao_c_out_io_buffer_reset.restype = None

        # Extra methods
        self._import_commands(interface_commands)
        self._import_commands(extra_commands)

        try:
            self.register_cell_magic()
        except:
            pass

        if init:
            self.init(init)
            
            
    def _import_commands(self, module):
        deny_list = getattr(module, '__deny_list', [])
        # Add in methods from `interface_commands`
        methods = [m for m in dir(module) if not m.startswith('__') and m not in deny_list]
        for m in methods:
            func = module.__dict__[m]
            setattr(self, m, types.MethodType(func, self))            

    #---------------------------------------------
    # Used by init and cmd routines

    def get_output(self, reset=True):
        """
        Returns a list of output strings.
        If reset, the internal Tao buffers will be reset.
        """
        n_lines = self.so_lib.tao_c_out_io_buffer_num_lines()
        lines = [self.so_lib.tao_c_out_io_buffer_get_line(i).decode('utf-8') for i in range(1, n_lines+1)]
        if reset:
            self.so_lib.tao_c_out_io_buffer_reset()
        return lines
    
    def reset_output(self):
        """
        Resets all output buffers
        """
        self.so_lib.tao_c_out_io_buffer_reset()

    #---------------------------------------------
    # Init Tao

    def init(self, cmd):

        if not tao_ctypes.initialized:
            logger.debug(f'Initializing Tao with: {cmd}')
            err = self.so_lib.tao_c_init_tao(cmd.encode('utf-8'))
            if err != 0:
                raise ValueError(f'Unable to init Tao with: {cmd}')
            tao_ctypes.initialized = True
            return self.get_output()
        else:
            # Reinit
            return self.cmd(f'reinit tao -clear {cmd}', raises=True)

    #---------------------------------------------
    # Send a command to Tao and return the output

    def cmd(self, cmd, raises=True):
        """
        Runs a command, and returns the text output
        
        cmd: command string
        raises: will raise an exception of [ERROR or [FATAL is detected in the output
        
        Returns a list of strings
        """
        
        logger.debug(f'Tao> {cmd}')

        self.so_lib.tao_c_command(cmd.encode('utf-8'))
        lines = self.get_output()
        
        # Error checking
        if not raises:
            return lines
        
        err = error_in_lines(lines)
        if err:
            raise RuntimeError(f'Command: {cmd} causes error: {err}')
        
        return lines
    
    def cmds(self, cmds, 
             suppress_lattice_calc=True, 
             suppress_plotting=True, 
             raises=True):
        """
        Runs a list of commands
    
        Args:
            cmds: list of commands
            
            suppress_lattice_calc: bool, optional
                If True, will suppress lattice calc when applying the commands
                Default: True
                
            suppress_plotting: bool, optional
                If True, will suppress plotting when applying commands
                Default: True
                
            raises: bool, optional
                If True will raise an exception of [ERROR or [FATAL is detected in the 
                output
                Default: True
            
        Returns:
            list of results corresponding to the commands
        
        """
        # Get globals to detect plotting
        g = self.tao_global()
        ploton, laton = g['plot_on'], g['lattice_calc_on']
        
        if suppress_plotting and ploton:
            self.cmd('set global plot_on = F')
        if suppress_lattice_calc and laton:
            self.cmd('set global lattice_calc_on = F')            
    
        # Actually apply commands
        results = []
        for cmd in cmds:
            res = self.cmd(cmd, raises=raises)
            results.append(res)
            
        if suppress_plotting and ploton:
            self.cmd('set global plot_on = T')
        if suppress_lattice_calc and laton:
            self.cmd('set global lattice_calc_on = T')               
            
        return results
            
        
    
    #---------------------------------------------
    # Get real array output.
    # Only python commands that load the real array buffer can be used with this method.

    def cmd_real (self, cmd, raises=True):
        logger.debug(f'Tao> {cmd}')
        
        self.so_lib.tao_c_command(cmd.encode('utf-8'))
        n = self.so_lib.tao_c_real_array_size()
        self.so_lib.tao_c_get_real_array.restype = ctypes.POINTER(ctypes.c_double * n)

        # Check the output for errors
        lines = self.get_output(reset=False)
        err = error_in_lines(lines)
        if err:
            self.reset_output()
            if raises:
                raise RuntimeError(err)
            else:
                return None
    
        # Extract array data
        # This is a pointer to the scratch space.
        array = np.ctypeslib.as_array(
            (ctypes.c_double * n).from_address(ctypes.addressof(self.so_lib.tao_c_get_real_array().contents)))
        
        array = array.copy()
        self.reset_output()
        
        return array  

    #----------
    # Get integer array output.
    # Only python commands that load the integer array buffer can be used with this method.

    def cmd_integer (self, cmd, raises=True):
        logger.debug(f'Tao> {cmd}')
        
        self.so_lib.tao_c_command(cmd.encode('utf-8'))
        n = self.so_lib.tao_c_integer_array_size()
        self.so_lib.tao_c_get_integer_array.restype = ctypes.POINTER(ctypes.c_int * n)

        # Check the output for errors
        lines = self.get_output(reset=False)
        err = error_in_lines(lines)
        if err:
            self.reset_output()
            if raises:
                raise RuntimeError(err)
            else:
                return None  
        
        # Extract array data
        # This is a pointer to the scratch space.
        array = np.ctypeslib.as_array(
            (ctypes.c_int * n).from_address(ctypes.addressof(self.so_lib.tao_c_get_integer_array().contents)))

        array = array.copy()
        self.reset_output()
        
        return array  
    
 

    #---------------------------------------------

    def register_cell_magic(self):
      """
      Registers a cell magic in Jupyter notebooks
      Invoke by
      %%tao
      sho lat
      """

      from IPython.core.magic import register_cell_magic
      @register_cell_magic
      def tao(line, cell):
          cell = cell.format(**globals())
          cmds=cell.split('\n')
          output = []
          for c in cmds:
              print('-------------------------')
              print('Tao> '+c)
              res = self.cmd(c)
              for l in res:
                   print(l)
      del tao


def find_libtao(base_dir):  
    """
    Searches base_for for an appropriate libtao shared library. 
    """
    for lib in ['libtao.so', 'libtao.dylib', 'libtao.dll']:
        so_lib_file = os.path.join(base_dir, lib)
        if os.path.exists(so_lib_file):
            return so_lib_file
    return None
    

def auto_discovery_libtao():
    """
    Use system loader to try and find libtao.
    """
    for lib in ['libtao.so', 'libtao.dylib', 'libtao.dll']:
        try:
            lib_handler = ctypes.CDLL(lib)
            return lib_handler, lib
        except OSError:
            continue
    return None, None


#----------------------------------------------------------------------

class TaoModel(Tao):
    """
    Base class for setting up a Tao model in a directory. Builds upon the Tao class.

    If use_tempdir==True, then the input_file and its directory will be copied to a temporary directory.
    If workdir is given, then this temporary directory will be placed in workdir.
    """

    def __init__(self,
          input_file='tao.init',
          ploton = True,
          use_tempdir=True,
          workdir=None,
          verbose=True,
          so_lib='',  # Passed onto Tao superclass
          auto_configure = True # Should be disables if inheriting.
          ):

        # NOTE: SUPER is being called from configure(...)

        # Save init

        self.original_input_file = input_file
        self.ploton = ploton
        self.use_tempdir = use_tempdir
        self.workdir = workdir
        if workdir: assert os.path.exists(workdir), 'workdir does not exist: '+workdir

        self.verbose=verbose
        self.so_lib=so_lib

        # Run control
        self.finished = False
        self.configured = False

        if os.path.exists(os.path.expandvars(input_file)):
            f = full_path(input_file)
            self.original_path, self.original_input_file = os.path.split(f) # Get original path, filename
            if auto_configure:
                self.configure()
        else:
            self.vprint('Warning: Input file does not exist. Cannot configure.')

    def configure(self):

        # Set paths
        if self.use_tempdir:
            # Need to attach this to the object. Otherwise it will go out of scope.
            self.tempdir = tempfile.TemporaryDirectory(dir=self.workdir)
            # Make yet another directory to overcome the limitations of shutil.copytree
            self.path = full_path(os.path.join(self.tempdir.name, 'tao/'))
            # Copy everything in original_path
            shutil.copytree(self.original_path, self.path, symlinks=True)
        else:
            # Work in place
            self.path = self.original_path

        self.input_file = os.path.join(self.path, self.original_input_file)

        self.vprint('Initialized Tao with '+self.input_file)


        # Set up Tao library
        super().__init__(init=self.init_line(), so_lib=self.so_lib)

        self.configured = True

    def init_line(self):
        line = '-init '+self.input_file
        if self.ploton:
            line += ' --noplot'
        else:
            line += ' -noplot'
        return line

    def reinit(self):
        line = 'reinit tao '+self.init_line()
        self.cmd(line)
        self.vprint('Re-initialized with '+line)

    def vprint(self, *args, **kwargs):
        # Verbose print
        if self.verbose:
            print(*args, **kwargs)
            
    #---------------------------------
    # Conveniences        

    @property
    def globals(self):
        """
        Returns dict of tao parameters.
        Note that the name of this function cannot be named 'global'
        """

        dat = self.cmd('python global')
        return tao_parameter_dict(dat)            

    #---------------------------------
    # [] for set command

    def __setitem__(self, key, item):
        """
        Issue a set command separated by :
        
        Example:
            TaoModel['global:track_type'] = 'beam'
        will issue command:
            set global track_type = beam
        """

        cmd = form_set_command(key, item,  delim=':')
        self.vprint(cmd)
        self.cmd(cmd)

    #---------------------------------
    def evaluate(self, expression):
        """
        Example: 
            .evaluate('lat::orbit.x[beginning:end]')
        Returns an np.array of floats
        """

        return tao_object_evaluate(self, expression)

    #---------------------------------
    def __str__(self):
        s = 'Tao Model initialized from: '+self.original_path
        s +='\n Working in path: '+self.path
        return s
        
#------------------------------------------------------------------------------- 
#------------------------------------------------------------------------------- 
# Helper functions        
     
def tao_object_evaluate(tao_object, expression):
    """
    Evaluates an expression and returns 
    
    Example expressions:
        beam::norm_emit.x[end]        # returns a single float
        lat::orbit.x[beginning:end]   # returns an np array of floats 
    """
    
    cmd = f'python evaluate {expression}'
    res = tao_object.cmd(cmd)

    # Cast to float
    vals = [x.split(';')[1] for x in res]
    
    try:
        fvals = np.asarray(vals, dtype=np.float)
    except:
        fvals = vals
    
    # Return single value, or array
    if len(fvals) == 1:
        return fvals[0]
    return fvals        
        
def form_set_command(s, value,  delim=':'):    
    """
    Forms a set command string that is separated by delim.
    
    Splits into three parts:
    command:what:attribute
    
    If 'what' had delim inside, the comma should preserve that.
    
    Example:
    >>>form_set_command('ele:BEG:END:a', 1.23)
    'set ele BEG:END a = 1.23'
    
    """
    x = s.split(delim)
    
    cmd0 = x[0]
    what = ':'.join(x[1:-1])
    att = x[-1]
    cmd = f'set {cmd0} {what} {att} = {value}'
    
    # cmd = 'set '+' '.join(x) + f' = {value}'
  
    return cmd
    
    
def apply_settings(tao_object, settings):
    """
    Applies multiple settings to a tao object.
    Checks for lattice_calc_on and plot_on, and temporarily disables these for speed.
    """
    
    cmds = []
    
    # Save these
    plot_on = tao_object.globals['plot_on'].value
    lattice_calc_on = tao_object.globals['lattice_calc_on'].value
    
    if plot_on:
        cmds.append('set global plot_on = F')
    if lattice_calc_on:
        cmds.append('set global lattice_calc_on = F')
    
    
    for k, v in settings.items():
        cmd = form_set_command(k, v)
        cmds.append(cmd)
        
    # Restore
    if lattice_calc_on:
        cmds.append('set global lattice_calc_on = T')

    if plot_on:
        cmds.append('set global plot_on = T')    
        
        
    for cmd in cmds:
        tao_object.vprint(cmd)
        tao_object.cmd(cmd, raises=True)
        
    return cmds    

#------------------------------------------------------------------------------- 
#------------------------------------------------------------------------------- 
# Helper functions        

def run_tao(settings=None,
                run_commands=['set global track_type=single'],
                input_file='tao.init',
                ploton=False,
                workdir=None,
                so_lib='',
                verbose=False):
    """
    Creates an LCLSTaoModel object, applies settings, and runs the beam.
    """
    
    assert os.path.exists(input_file), f'Tao input file does not exist: {input_file}'
    
    M = TaoModel(input_file=input_file,
                 ploton = ploton,
                 use_tempdir=True,
                 workdir=workdir,
                 verbose=verbose,
                 so_lib=so_lib,  # Passed onto Tao superclass
                 auto_configure = True) # Should be disables if inheriting.
                
    # Move to local dir, so call commands work 
    init_dir = os.getcwd()
    os.chdir(M.path)
    
    try:
        if settings:
            apply_settings(M, settings)
        
        for command in run_commands:
            if verbose:
                print('run command:', command)
            M.cmd(command, raises=True)
    
    finally:
        # Return to init_dir
        os.chdir(init_dir)    

    return M



data_type_list = []
data_type_list.append({'name': '', 'data_source': ['beam', 'lat'], 's_offset':True})
data_type_list.append({'name': 'alpha.<enum>', '<enum>': ['a', 'b', 'z'], 'data_source': ['lat'], 's_offset':True})
data_type_list.append({'name': 'apparent_emit.<enum>', '<enum>': ['x', 'y'], 'data_source': ['beam', 'lat'], 's_offset':False})
data_type_list.append({'name': 'norm_apparent_emit.<enum>', '<enum>': ['x', 'y'], 'data_source': ['beam', 'lat'], 's_offset':False})
data_type_list.append({'name': 'beta.<enum>', '<enum>': ['x', 'y', 'z', 'a', 'b', 'c'], 'data_source': ['beam', 'lat'], 's_offset':True})
data_type_list.append({'name': 'bpm_orbit.<enum>', '<enum>': ['x', 'y'], 'data_source': ['lat'], 's_offset':True})
data_type_list.append({'name': 'bpm_eta.<enum>', '<enum>': ['x', 'y'], 'data_source': ['lat'], 's_offset':True})
data_type_list.append({'name': 'bpm_phase.<enum>', '<enum>': ['a', 'b'], 'data_source': ['lat'], 's_offset':True})
data_type_list.append({'name': 'bpm_k.<enum>', '<enum>': ['22a', '12a', '11b', '12b'], 'data_source': ['lat'], 's_offset':True})
data_type_list.append({'name': 'bpm_cbar.<enum>', '<enum>': ['22a', '12a', '11b', '12b'], 'data_source': ['lat'], 's_offset':True})
data_type_list.append({'name': 'bunch_min.<enum>', '<enum>': ['x', 'px', 'y', 'py', 'z', 'pz'], 'data_source': ['beam'], 's_offset':False})
data_type_list.append({'name': 'bunch_max.<enum>', '<enum>': ['x', 'px', 'y', 'py', 'z', 'pz'], 'data_source': ['beam'], 's_offset':False})
data_type_list.append({'name': 'bunch_charge.<enum>', '<enum>': ['live', 'live_relative'], 'data_source': ['beam'], 's_offset':False})
data_type_list.append({'name': 'c_mat.<enum>', '<enum>': ['11', '12', '21', '22'], 'data_source': ['lat'], 's_offset':True})
data_type_list.append({'name': 'cbar.<enum>', '<enum>': ['11', '12', '21', '22'], 'data_source': ['lat'], 's_offset':True})
data_type_list.append({'name': 'chrom.<enum>', '<enum>': ['dtune.a', 'a', 'dtune.b', 'b', 'dbeta.a', 'dbeta.b', 'dphi.a', 'dphi.b', 'deta.x', 'deta.y', 'detap.x', 'detap.y'], 'data_source': ['lat'], 's_offset':False})
data_type_list.append({'name': 'damp.<enum>', '<enum>': ['j_a', 'j_b', 'j_z'], 'data_source': ['lat'], 's_offset':False})
data_type_list.append({'name': 'dpx_dx', 'data_source': ['beam'], 's_offset':False})
data_type_list.append({'name': 'dpy_dy', 'data_source': ['beam'], 's_offset':False})
data_type_list.append({'name': 'dpz_dz', 'data_source': ['beam'], 's_offset':False})
data_type_list.append({'name': 'e_tot', 'data_source': ['lat'], 's_offset':False})
data_type_list.append({'name': 'e_tot_ref', 'data_source': ['lat'], 's_offset':False})
data_type_list.append({'name': 'element_attrib.<str>', 'data_source': ['lat'], 's_offset':False})
data_type_list.append({'name': 'emit.<enum>', '<enum>': ['x', 'y', 'z', 'a', 'b', 'c'], 'data_source': ['beam', 'lat'], 's_offset':False})
data_type_list.append({'name': 'norm_emit.<enum>', '<enum>': ['x', 'y', 'z', 'a', 'b', 'c'], 'data_source': ['beam', 'lat'], 's_offset':False})
data_type_list.append({'name': 'eta.<enum>', '<enum>': ['a', 'b', 'x', 'y', 'z'], 'data_source': ['beam', 'lat'], 's_offset':True})
data_type_list.append({'name': 'etap.<enum>', '<enum>': ['a', 'b', 'x', 'y'], 'data_source': ['beam', 'lat'], 's_offset':True})
data_type_list.append({'name': 'expression.<str>', 'data_source': ['lat'], 's_offset':False})
data_type_list.append({'name': 'floor.<enum>', '<enum>': ['x', 'y', 'z', 'theta', 'phi', 'psi'], 'data_source': ['lat'], 's_offset':True})
data_type_list.append({'name': 'floor_actual.<enum>', '<enum>': ['x', 'y', 'z', 'theta', 'phi', 'psi'], 'data_source': ['lat'], 's_offset':True})
data_type_list.append({'name': 'floor_orbit.<enum>', '<enum>': ['x', 'y', 'z'], 'data_source': ['beam', 'lat'], 's_offset':True})
data_type_list.append({'name': 'gamma.<enum>', '<enum>': ['a', 'b', 'z'], 'data_source': ['lat'], 's_offset':True})
data_type_list.append({'name': 'k.<enum>', '<enum>': ['11b', '12a', '12b', '22a'], 'data_source': ['lat'], 's_offset':True})
data_type_list.append({'name': 'srdt.<enum1>.<enum2>',
    '<enum1>': ['h20001', 'h00201', 'h10002', 'h21000', 'h30000', 'h10110', 'h10020', 'h10200', 'h31000', 'h40000', 'h20110', 'h11200', 'h20020', 'h20200', 'h00310', 'h00400', 'h22000', 'h00220', 'h11110'],
    '<enum2>': ['r', 'i', 'a'], 'data_source': ['lat'], 's_offset':False})
data_type_list.append({'name': 'normal.h.<enum1>.<enum2>',
    '<enum1>': ['20001', '00201', '10002', '21000', '30000', '10110', '10020', '10200', '31000', '40000', '20110', '11200', '20020', '20200', '00310', '00400', '22000', '00220', '11110'],
    '<enum2>': ['r', 'i', 'a'], 'data_source': ['lat'], 's_offset':False})
data_type_list.append({'name': 'normal.<enum>.<digit:1-6>.<digits6>', '<enum>': ['M', 'A', 'A_inv', 'dhdj', 'RaF', 'ImF'], 'data_source': ['lat'], 's_offset':False})
data_type_list.append({'name': 'momentum', 'data_source': ['lat'], 's_offset':True})
data_type_list.append({'name': 'momentum_compaction', 'data_source': ['lat'], 's_offset':False})
data_type_list.append({'name': 'n_particle_loss', 'data_source': ['beam'], 's_offset':False})
data_type_list.append({'name': 'orbit.<enum>', '<enum>': ['e_tot', 'x', 'y', 'z', 'px', 'py', 'pz', 'amp_a', 'amp_b', 'norm_amp_a', 'norm_amp_b'], 'data_source': ['beam', 'lat'], 's_offset':True})
data_type_list.append({'name': 'pc', 'data_source': ['beam', 'lat'], 's_offset':False})
data_type_list.append({'name': 'periodic.tt.<int>', 'data_source': ['lat'], 's_offset':False})
data_type_list.append({'name': 'phase.<enum>', '<enum>': ['a', 'b'], 'data_source': ['lat'], 's_offset':True})
data_type_list.append({'name': 'phase_frac.<enum>', '<enum>': ['a', 'b'], 'data_source': ['lat'], 's_offset':False})
data_type_list.append({'name': 'phase_frac_diff', 'data_source': ['lat'], 's_offset':False})
data_type_list.append({'name': 'photon.<enum>', '<enum>': ['intensity_x', 'intensity_y', 'intensity', 'phase_x', 'phase_y'], 'data_source': ['beam', 'lat'], 's_offset':False})
data_type_list.append({'name': 'ping_a.<enum>', '<enum>': ['amp_x', 'phase_x', 'amp_y', 'phase_y', 'amp_sin_y', 'amp_cos_y', 'amp_sin_rel_y', 'amp_cos_rel_y'], 'data_source': ['lat'], 's_offset':False})
data_type_list.append({'name': 'ping_b.<enum>', '<enum>': ['amp_y', 'phase_y', 'amp_x', 'phase_x', 'amp_sin_x', 'amp_cos_x', 'amp_sin_rel_x', 'amp_cos_rel_x'], 'data_source': ['lat'], 's_offset':False})
data_type_list.append({'name': 'r.<digit:1-6><digit:1-6>', 'data_source': ['lat'], 's_offset':False})
data_type_list.append({'name': 'r56_compaction', 'data_source': ['lat'], 's_offset':False})
data_type_list.append({'name': 'rad_int.<enum>', '<enum>': ['i0', 'i1', 'i2', 'i2_e4', 'i3', 'i3_e7', 'i4a', 'i4b', 'i4z', 'i5a', 'i5a_e6', 'i5b', 'i5b_e6', 'i6b'], 'data_source': ['lat'], 's_offset':False})
data_type_list.append({'name': 'rad_int1.<enum>', '<enum>': ['i0', 'i1', 'i2', 'i2_e4', 'i3', 'i3_e7', 'i4a', 'i5a', 'i5a_e6', 'i4b', 'i5b', 'i5b_e6', 'i6b'], 'data_source': ['lat'], 's_offset':False})
data_type_list.append({'name': 'ref_time', 'data_source': ['beam', 'lat'], 's_offset':True})
data_type_list.append({'name': 'rel_floor.<enum>', '<enum>': ['x', 'y', 'z', 'theta', 'phi', 'psi'], 'data_source': ['lat'], 's_offset':False})
data_type_list.append({'name': 'sigma.<enum>', '<enum>': ['x', 'px', 'y', 'py', 'z', 'pz', 'xy', 'Lxy'], 'data_source': ['beam', 'lat'], 's_offset':False})
data_type_list.append({'name': 'spin.<enum>', '<enum>': ['x', 'y', 'z', 'amp'], 'data_source': ['beam', 'lat'], 's_offset':False})
data_type_list.append({'name': 'spin.depolarization_rate', 'data_source': ['lat'], 's_offset':False})
data_type_list.append({'name': 'spin.polarization_rate', 'data_source': ['lat'], 's_offset':False})
data_type_list.append({'name': 'spin.polarization_limit', 'data_source': ['lat'], 's_offset':False})
data_type_list.append({'name': 'spin_g_matrix.<enum>', '<enum>': ['11', '12', '13', '14', '15', '16', '21', '22', '23', '24', '25', '26'], 'data_source': ['lat'], 's_offset':False})
data_type_list.append({'name': 's_position', 'data_source': ['lat'], 's_offset':True})
data_type_list.append({'name': 'time', 'data_source': ['beam', 'lat'], 's_offset':True})
data_type_list.append({'name': 'tune.<enum>', '<enum>': ['a', 'b'], 'data_source': ['lat'], 's_offset':False})
data_type_list.append({'name': 't.<int>', 'data_source': ['lat'], 's_offset':False})
data_type_list.append({'name': 'tt.<int>', 'data_source': ['lat'], 's_offset':False})
data_type_list.append({'name': 'unstable.<enum>', '<enum>': ['orbit', 'ring'], 'data_source': ['lat'], 's_offset':False})
data_type_list.append({'name': 'velocity.<enum>', '<enum>': ['', 'x', 'y', 'z'], 'data_source': ['beam', 'lat'], 's_offset':True})
data_type_list.append({'name': 'wall.<enum>', '<enum>': ['left_side', 'right_side'], 'data_source': ['lat'], 's_offset':False})
data_type_list.append({'name': 'wire.<real>', 'data_source': ['beam'], 's_offset':False})



from .csr import csr_wake_stats_at_step

import matplotlib.pyplot as plt
import numpy as np


def plot_csr_wake(data, step=0, **kwargs):
    """
    Plots a step from process CSR wake data
    
    
    A simple interactive plot can be made with:
    
        from ipywidgets import interact
        def plot1(step=0):
            plot_csr_wake(data, step=step)
        nstep = len(data['z'])
        interact(plot1, step=(0, nstep-1, 1) )
    
    """
    
    fig, ax = plt.subplots(**kwargs)

    
    z = data['z'][step,:]
    density =  data['charge_per_meter'][step,:]
    kick = data['csr_kick_per_meter'][step,:]
    
    # Full bounds for plot
    zmin = data['z'].min()
    zmax = data['z'].max()
    
    kmin = np.min(data['csr_kick_per_meter'])
    kmax = np.max(data['csr_kick_per_meter'])
    
    #nz = len(z)
    #dz = z.ptp()/(nz-1)
    #dsum = np.sum(density)
    #qtot = dsum*dz 
    ## Normalize for stats
    #density = density/dsum
            
    ax.set_xlabel('z (µm)')
    ax.set_ylabel('CSR Kick/m')
    
    
    
    zscale = 1e6 # m -> µm
    ax.set_xlim(zmin*zscale, zmax*zscale)
    ax.set_ylim(kmin, kmax)
    
   
    ax2 = ax.twinx()
    ax2.set_ylabel('density')
    ax2.fill(z*zscale, density, color='grey', alpha=0.5)
    
    ax.plot(z*zscale, kick, color='blue')
    
    
    
def plot_csr_stats(data, **kwargs):
    """
    Plots csr wake stats along s from processed csr_wake data
    """
    
    s_pos = data['s_position']
    nsteps = len(data['z'])
    
    stats = np.array([csr_wake_stats_at_step(data, step=i) for i in range(nsteps)])
    
    fig, ax = plt.subplots(**kwargs)
    
    ax.set_xlabel('s (m)')
    ax.set_ylabel('CSR Kick/m')
    ax.plot(s_pos, stats[:,0], color='black', label='Average Wake')
    ax.plot(s_pos, stats[:,1], color='red', label='std Wake')
    
    ax2 = ax.twinx()
    ax2.plot(s_pos, stats[:,3]*1e15/299792458, color='blue', label='$\sigma_z/c (fs)$')
    ax2.set_ylabel('$\sigma_z/c (fs)$')
    ax.legend()

    
    
 



'''
This module defines a modified version of the tao_interface class with
a few extra features to facilitate communication with the GUI
'''
import tkinter as tk
import time

import pytao
from pytao.tao_interface import new_stdout, filter_output

class tao_interface(pytao.tao_interface):
    '''
    Provides an interface between the GUI and
    the Tao command line
    '''
    def __init__(self, mode="ctypes", init_args = "", tao_exe =  "", expect_str = "Tao>", so_lib=""):
        # DEBUG
        self.debug = False # set true to print debug messages
        # Put patterns of interest here
        self.debug_patterns = ['python plot_manage']
        if self.debug:
            print(init_args)
        ###
        pytao.tao_interface.__init__(
                self, mode=mode, init_args=init_args, tao_exe=tao_exe, expect_str=expect_str, so_lib=so_lib)
        self.message = ""
        self.message_type = "" #conveys color info to console
        self.printed = tk.BooleanVar()
        self.printed.set(False)
        # DEV PURPOSES ONLY
        self._root = self.printed._root

    def cmd_in(self, cmd_str, no_warn=False):
        '''
        Runs cmd_str at the Tao command line and returns the as a string
        Warning messages generated by this command are suppressed when
        no_warn is set True
        '''
        # DEBUG
        if self.debug:
            match = False
            if self.debug_patterns == 'All':
                match = True
            elif isinstance(self.debug_patterns, list):
                for p in self.debug_patterns:
                    if cmd_str.find(p) == 0:
                        match = True
            if match:
                current_time = time.localtime()
                y = str(current_time.tm_year)
                mo = str(current_time.tm_mon)
                if len(mo) == 1:
                    mo = '0' + mo
                day = str(current_time.tm_mday)
                if len(day) == 1:
                    day = '0' + day
                hr = str(current_time.tm_hour)
                if len(hr) == 1:
                    hr = '0' + hr
                mn = str(current_time.tm_min)
                if len(mn) == 1:
                    mn = '0' + mn
                sec = str(current_time.tm_sec)
                if len(sec) == 1:
                    sec = '0' + sec
                print(y+'-'+mo+'-'+day+' '+hr+':'+mn+':'+sec, cmd_str)
        if cmd_str.find("dev ") ==0:
            cmd_str = cmd_str[4:]
            with new_stdout() as output:
                exec(cmd_str)
            output = output.getvalue()
            output = filter_output(output)
            return output
        ###

        self.message = ""
        self.message_type = "normal" #conveys color info to console
        if cmd_str.find("single_mode") == 0:
            output = "single_mode not supported on the GUI command line"
            self.message_type = "error"
        else:
            output = pytao.tao_interface.cmd_in(self, cmd_str)

        # Parse output for error messages
        if cmd_str.find("spawn ") == 0:
            self.message += "WARNING: the spawn command is very dangerous "
            self.message += "and may cause the GUI to hang if the spawned "
            self.message += "process does not exit quickly."
            self.message_type = "error"
            self.printed.set(True)
            self.message = ""
            self.message_type = "normal"
        if output.find("[ERROR") != -1:
            self.message += "Warning: Error occurred in Tao\n"
            self.message += "The offending command: " + cmd_str + "\n"
            self.message += "The error (first 20 lines):\n"
            if len(output.splitlines()) > 20:
                for i in range(20):
                    self.message += (output.splitlines()[i]) + "\n"
            else:
                self.message += output
            self.message_type = "error"
            if not no_warn:
                self.printed.set(True)
        if (output.find("Backtrace") != -1) | \
                (output.find("SIGSEGV") != -1):
            self.message = "Error occurred in Tao, causing it to crash\n"
            self.message += "The offending command: " + cmd_str + '\n'
            self.message += "The error:\n"
            self.message += output + '\n'
            self.message_type = "fatal"
            if not no_warn:
                self.printed.set(True)
        return output

    def cmd(self, cmd_str):
        '''
        Runs cmd_str at the Tao command line and prints the output
        '''
        output = self.cmd_in(cmd_str, no_warn=True)
        self.message_type = "normal"
        self.message = output
        self.printed.set(True)



import pkgutil
import sys
def module_check():
    '''
    Checks for the required modules (listed in the required list).
    If any modules are missing, information on how to install the missing modules
    is printed and execution is halted.
    '''
    missing = [] # Missing modules
    upgrade = False
    required = ["tkinter",
            "pexpect",
            "matplotlib"]
    for mod in required:
        if pkgutil.find_loader(mod) == None:
            missing.append(mod)
    # Check matplotlib version
    if "matplotlib" not in missing:
        import matplotlib
        v = matplotlib.__version__
        if v[0] != '3':
            print("Warning: matplotlib version 3 or higher is required"
                    + " (your version:" + v + ")")
            upgrade = True

    if missing != []:
        print("Warning: some required python modules were not found.")
        print("Missing modules:")
        for mod in missing:
            print(mod)
        print("Please install these modules using your system package manager,"
                + "pip, macports, or whatever you use to manager your python modules")
        print("Linux system package managers:")
        print("      sudo apt-get install python-" + mod)
        print("      sudo pacman -S python-" + mod)
        print("Using pip:")
        print("      pip install " + missing[0])
        print("Using macports (Mac users):")
        print("      sudo port install py36-tkinter")
        print("Note: it is generally a bad idea to use sudo pip install,"
                + " as this may break parts of your system python installation"
                + " that the os depends on to function properly.    "
                + "It is a better idea to set up a virtual environment "
                + "and install your modules there.")
        print("Note: it is a good idea to search for the exact name of the "
                + "modules you need before trying to install them.  "
                + "Use the following commands to do so.")
        print("Debian based distros (incl. Ubuntu):")
        print("      apt-search " + mod)
        print("Arch based distros (incl. Manjaro):")
        print("      pacman -Ss " + mod)
        print("Using pip:")
        print("      pip search " + missing[0])
        print("Using macports (Mac users):")
        print("      port search tkinter")
        sys.exit()
    if upgrade:
        sys.exit()



'''
Provides windows for viewing and editing variable arrays in tao
'''
import tkinter as tk
from tkinter import ttk
from tkinter import messagebox
from tkinter import filedialog

from pytao.util.parameters import str_to_tao_param, tao_parameter_dict
from .tao_widget import *
from .tao_set import *
from .tao_base_windows import *
from .tao_lat_windows import tao_ele_browser
#-----------------------------------------------------
# Variable Window

class tao_var_general_window(tao_list_window):

    def __init__(self, root, pipe, *args, **kwargs):
        self.tao_id = 'var'
        tao_list_window.__init__(self, root, "v1 Variables", *args, **kwargs)
        self.pipe = pipe
        for i in [0,4,5]:
            self.list_frame.grid_columnconfigure(i, pad=10)
        self.refresh()

    def refresh(self):
        for child in self.list_frame.winfo_children():
            child.destroy()
        v1_var_list = self.pipe.cmd_in("python var_general")
        v1_var_list = v1_var_list.splitlines()
        # close if there are no variables
        if len(v1_var_list) == 0:
            messagebox.showwarning('Warning', 'No variables defined', parent=self)
            self.destroy()
            return
        for i in range(len(v1_var_list)):
            v1_var_list[i] = v1_var_list[i].split(';')

        tk.Label(self.list_frame, text="Variable").grid(
                row=0, column=0, columnspan=4, sticky='W')
        tk.Label(self.list_frame, text="Indices").grid(row=0, column=4)
        tk.Label(self.list_frame, text="Using").grid(row=0, column=5)

        i=1
        for item in v1_var_list:
            tk.Label(self.list_frame, text=item[0]).grid(row=i, column=0, sticky='W')
            tk.Button(self.list_frame, text="View...",
                    command=self.open_v1_callback(item[0])).grid(row=i, column=1)
            tk.Button(self.list_frame, text="Edit...",
                    command=self.edit_v1_callback(item[0])).grid(row=i, column=2)
            tk.Button(self.list_frame, text="Delete...",
                    command=self.delete_v1_callback(item[0])).grid(row=i, column=3)
            #tk.Button(self.list_frame, text="Write...",
            #        command=self.write_v1).grid(row=i, column=3)
            tk.Label(self.list_frame,text=item[2] +':'+ item[3]).grid(row=i,column=4)
            tk.Label(self.list_frame, text=item[1]).grid(row=i, column=5)
            i = i+1

    def open_v1_callback(self, v1_var_name):
        return lambda : self.open_v1(v1_var_name)

    def open_v1(self, v1_var_name):
        win = tao_v1_var_window(self.root, self.pipe, v1_var_name)

    def edit_v1_callback(self, v1_var_name):
        return lambda : self.edit_v1(v1_var_name)

    def edit_v1(self, v1_var_name):
        win = tao_new_var_window(self.root, self.pipe, default=v1_var_name)

    def delete_v1_callback(self, v1_var_name):
        return lambda : self.delete_v1(v1_var_name)

    def delete_v1(self, v1_var_name):
        ans = messagebox.askokcancel(title="Delete " + v1_var_name + "?",
                message="Are you sure you want to delete " + v1_var_name
                + "?")
        if ans:
            self.pipe.cmd_in("python var_v1_destroy " + v1_var_name)
            for win in self.root.refresh_windows['var']:
                win.refresh()
            for win in self.root.refresh_windows['plot']:
                win.refresh()


#-----------------------------------------------------
# v1_var window

class tao_v1_var_window(lw_table_window):

    def __init__(self, root, pipe, v1_var_name, *args, **kwargs):
        self.tao_id = 'var'
        title_list = ["Index",
                "Name",
                "Meas value",
                "Model value",
                "Design value",
                "Useit_opt",
                "good_user",
                "Weight"]
        bulk_template = []
        self.lw = root.lw #set light-weight flag
        bulk_template.append([str_to_tao_param("meas_value;REAL;T;"), 2])
        bulk_template.append([str_to_tao_param("good_user;LOGIC;T;"), 6])
        bulk_template.append([str_to_tao_param("weight;REAL;T;"), 7])

        bulk_set_format = "set var {}|"
        set_format = "set var {}[{}]|"
        if self.lw:
            lw_table_window.__init__(self, root, pipe, v1_var_name, title_list,
                    bulk_template, bulk_set_format, set_format, *args, **kwargs)
        else:
            table_window.__init__(self, root, pipe, v1_var_name, title_list,
                    bulk_template, bulk_set_format, set_format, *args, **kwargs)

    def refresh(self):
        self.v_list = self.pipe.cmd_in("python var_v_array " + self.array_name)
        self.v_list = self.v_list.splitlines()
        self.row_num = len(self.v_list)
        if self.lw:
            lw_table_window.refresh(self)
        else:
            table_window.refresh(self)

    def list_row_fetch(self, index):
        '''
        Returns row number index to be appended to self.list_rows
        '''
        return v1_var_list_entry(self.list_frame, self.v_list[index])

    def lw_list_row_fetch(self, index):
        '''
        For use with lw_table_window.  Returns a list of the values
        to be displayed in each column of row #index.
        '''
        return self.v_list[index].split(';')

    def lw_detail_callback(self, event=None):
        '''
        Callback for the light-weight table to open a detail window
        '''
        lw_table_window.open_detail_window_callback(self, event)

    def bulk_set(self, fill_choices, parent):
        '''
        Overload of lw_table_window.bulk_set that calls refresh data and plot windows
        '''
        lw_table_window.bulk_set(self, fill_choices, parent)
        # Refresh data-related windows
        for win in self.root.refresh_windows['var']:
            win.refresh()
        for win in self.root.refresh_windows['plot']:
            win.refresh()

    def open_detail_window(self, index):
        self.param_list = self.pipe.cmd_in(
                "python var " + self.array_name + '[' + str(index) + ']')
        self.param_list = self.param_list.splitlines()
        for i in range(len(self.param_list)):
            self.param_list[i]=str_to_tao_param(self.param_list[i])
        if self.lw:
            lw_table_window.open_detail_window(self, index)
        else:
            table_window.open_detail_window(self, index)

    def detail_set_callback(self, tao_list, set_str):
        '''
        Overload of lw_table_window.detail_set_callback that refreshes
        data-related windows
        '''
        lw_table_window.detail_set_callback(self, tao_list, set_str)
        # Refresh data-related windows
        for win in self.root.refresh_windows['var']:
            win.refresh()
        for win in self.root.refresh_windows['plot']:
            win.refresh()

class tao_new_var_window(Tao_Toplevel):
    '''
    Provides a window for creating new v1_variable arrays
    '''
    def __init__(self, root, pipe, default=None, *args, **kwargs):
        self.root = root
        Tao_Toplevel.__init__(self, root, *args, **kwargs)
        self.pipe = pipe
        self.title('New Variables')
        self.v1_frame_list = []

        # Possibly create self.notebook
        self.notebook = ttk.Notebook(self)
        self.notebook.pack(side='top', fill='both', expand=1)

        # Set up a v1 tab
        self.v1_frame_list.append(new_v1_frame(self))
        self.notebook.insert('end', self.v1_frame_list[0])
        self.notebook.tab(0, text=self.v1_frame_list[0].BLANK_TITLE)
        self.v1_index = 0 #marks current tab index

        # New tab button
        self.new_tab_frame = tk.Frame(self.notebook)
        self.notebook.insert('end', self.new_tab_frame)
        self.notebook.tab(1, text='+')
        self.notebook.bind('<<NotebookTabChanged>>', self.tab_handler)

        # Create button
        self.create_b = tk.Button(self, text="Create!", command=self.create_variables)
        self.create_b.pack(side='right')

        # Clone default if it exists
        if default != None:
            self.v1_frame_list[0].clone(default)

    def create_variables(self, event=None):
        '''
        Takes the information from the variable frames and runs
        the necessary commands to create the variables in tao, then closes
        the create variable window
        '''
        # Input validation
        messages = []
        missing_atts = [] #variables with element must have an attribute
        for v1_frame in self.v1_frame_list:
            # Check names
            if not v1_frame.name_handler(strict=True):
                name_m = "Please check v1_array names."
                if name_m not in messages:
                    messages.append(name_m)
            # Check min indices
            if not v1_frame.ix_min_handler(strict=True):
                messages.append("Please check the start index for " + v1_frame.name)
            # Check lengths
            if not v1_frame.length_handler(strict=True):
                messages.append("Please check the end index for " + v1_frame.name)
            # Check low/high limits
            if not v1_frame.low_high_handler(strict=True):
                messages.append("Please check the low and high limits for " + v1_frame.name)
            # Check for semicolons in any fields
            semi_message = "Semicolons not allowed in any input field"
            caret_message = "Carets not allowed in any input field"
            space_message = "Spaces not allowed in any input field"
            broken = False #Used to break out of the below for loops
            # Check for semicolons/carets
            for ttp in v1_frame.v1_array_wids:
                if str(ttp.tk_var.get()).find(';') != -1:
                    messages.append(semi_message)
                    broken = True
                    break
                if str(ttp.tk_var.get()).find('^') != -1:
                    messages.append(caret_message)
                    broken = True
                    break
                if str(ttp.tk_var.get()).find(' ') != -1:
                    messages.append(space_message)
                    broken = True
                    break
            for var_dict in v1_frame.var_dict.values():
                if broken:
                    break
                for v in var_dict.values():
                    if str(v).find(';') != -1:
                        messages.append(semi_message)
                        broken = True
                        break
                    if str(v).find('^') != -1:
                        messages.append(caret_message)
                        broken = True
                        break
                    if str(ttp.tk_var.get()).find(' ') != -1:
                        messages.append(space_message)
                        broken = True
                        break
                if broken:
                    break
            # Check that variables with elements have attributes
            for ix, var_dict in v1_frame.var_dict.items():
                if 'ele_name' in var_dict.keys():
                    if var_dict['ele_name'] != "":
                        if 'attribute' in var_dict.keys():
                            if var_dict['attribute'] == "":
                                missing_atts.append(ix)
                        elif v1_frame.v1_array_wids[2].tk_var.get() == "":
                            missing_atts.append(ix)
        if missing_atts != []:
            missing_atts.sort()
            att_msg = "The following variables were assigned elements but not attributes:\n"
            for ix in missing_atts:
                att_msg += v1_frame.name + '[' + str(ix) + ']' + '\n'
            messages.append(att_msg)
        for m in messages:
            messagebox.showwarning("Error", m, parent=self)
        if messages != []:
            return
        for v1_frame in self.v1_frame_list:
            # Create the variable array
            cmd_str = 'python var_v1_create ' + v1_frame.name + ' '
            cmd_str += str(v1_frame.ix_min) + ' '
            cmd_str += str(v1_frame.ix_max)
            self.pipe.cmd_in(cmd_str)
            # Create the individual variables
            for i in range(v1_frame.ix_min, v1_frame.ix_max+1):
                cmd_str = 'python var_create ' + v1_frame.name
                cmd_str += '[' + str(i) + ']^^'
                if i in v1_frame.var_dict.keys():
                    var_dict = v1_frame.var_dict[i]
                else:
                    var_dict = {}
                params = ['ele_name', 'attribute', 'universes', 'weight', 'step',
                        'low_lim', 'high_lim', 'merit_type', 'good_user', 'key_bound',
                        'key_delta']
                # Used to look up v1-level default parameters
                v1_params = ['name', 'universes', 'attribute', 'weight', 'step', 'merit_type',
                        'low_lim', 'high_lim', 'good_user', 'key_bound', 'key_delta']
                for j in range(len(params)):
                    p = params[j]
                    if j != 0:
                        v1_ix = v1_params.index(p)
                    else:
                        v1_ix = 0
                    # Highest priority: individual var settings
                    if p in var_dict.keys():
                        if p == 'universes' and var_dict[p]=="": #cannot be empty
                            u = v1_frame.v1_array_wids[v1_ix].tk_var.get()
                            if u == "":
                                u = '*'
                            cmd_str += u + '^^'
                        else:
                            if v1_frame.v1_array_wids[v1_ix].param.type == 'LOGIC':
                                cmd_str += ('T^^' if var_dict[p] else 'F^^')
                            else:
                                cmd_str += str(var_dict[p]) + '^^'
                    else: #Use defaults if var parameter not set
                        if j == 0:
                            cmd_str += '^^'
                            continue
                        if v1_frame.v1_array_wids[v1_ix].param.type == 'LOGIC':
                            cmd_str += ('T^^' if v1_frame.v1_array_wids[v1_ix].tk_var.get() else 'F^^')
                        elif p == 'universes': #cannot be empty
                            u = v1_frame.v1_array_wids[v1_ix].tk_var.get()
                            if u == "":
                                u = '*'
                            cmd_str += u + '^^'
                        else:
                            cmd_str += v1_frame.v1_array_wids[v1_ix].tk_var.get() + '^^'
                cmd_str = cmd_str[:-2] # remove ending ^^
                self.pipe.cmd_in(cmd_str)
        # Close the window
        self.destroy()
        # Refresh variable-related windows
        for win in self.root.refresh_windows['var']:
            win.refresh()
        for win in self.root.refresh_windows['plot']:
            win.refresh()

    def tab_handler(self, event=None):
        '''
        Handles new tab creation and updates self.v1_index as necessary
        '''
        # Check if the new tab frame has been selected
        if self.notebook.select() == self.new_tab_frame._w:
            # Add new tab
            self.v1_frame_list.append(new_v1_frame(self))
            self.v1_index = len(self.v1_frame_list)-1
            self.notebook.insert(self.v1_index, self.v1_frame_list[-1])
            self.notebook.tab(self.v1_index, text=self.v1_frame_list[-1].BLANK_TITLE)
            self.notebook.select(self.v1_index)
        else:
            # Update self.v1_index
            for i in range(len(self.v1_frame_list)):
                frame = self.v1_frame_list[i]
                if self.notebook.select() == frame._w:
                    self.v1_index = i
                    # Unblock frame's handlers
                    frame.handler_block = False
                    break

    def refresh(self):
        '''
        Only here in case something tries to refresh this window
        '''
        pass


class new_v1_frame(tk.Frame):
    '''
    Provides a frame for inputting properties of a v1_variable array.
    '''
    def __init__(self, parent):
        tk.Frame.__init__(self, parent.notebook)
        self.parent = parent
        self.pipe = self.parent.pipe
        self.BLANK_TITLE = "[UNTITLED]"
        self.name = self.BLANK_TITLE #Default
        self.clone_of = "" # Used to track clones of existing v1_arrays
        self.handler_block = False

        # d1 Widgets
        def v1_ttp(x):
            ''' Shortcut for commonly used construct '''
            return tk_tao_parameter(str_to_tao_param(x), self, self.pipe)

        self.v1_array_wids = [v1_ttp('name;STR;T;'),
                v1_ttp('universes;STR;T;'),
                v1_ttp('attribute;STR;T;'),
                v1_ttp('weight;REAL;T;'),
                v1_ttp('step;REAL;T;'),
                v1_ttp('var^merit_type;ENUM;T;'),
                v1_ttp('low_lim;REAL;T;'),
                v1_ttp('high_lim;REAL;T;'),
                v1_ttp('good_user;LOGIC;T;T'),
                v1_ttp('key_bound;LOGIC;T;'),
                v1_ttp('key_delta;REAL;T;'),
                v1_ttp('ix_min;INT;T;'),
                v1_ttp('ix_max;INT;T;')]
        # v1 labels (NAMES AS STRINGS ONLY)
        self.v1_array_labels = ["v1_array Name:", "Default universe:", "Default attribute:",
                "Default weight:", "Default step:", "Default merit type:",
                "Default low_lim:", "Default high_lim:", "Default good_user:",
                "Default key_bound:", "Default key_delta:", "Base index:", "Array length:"]
        # Grid widgets and labels:
        for i in range(len(self.v1_array_wids)):
            tk.Label(self, text=self.v1_array_labels[i]).grid(row=i+3, column=0, sticky='W')
            self.v1_array_wids[i].tk_wid.grid(row=i+3, column=1, sticky='EW')
        i = i+3

        # Warning labels
        # (defined here to be gridded/ungridded as necessary)
        self.name_warning_empty = tk.Label(self, text="Must not be empty")
        self.name_warning_invalid = tk.Label(self, text="v1 name already in use")
        self.ix_min_warning = tk.Label(self, text="Must be a non-negative integer")
        self.length_warning = tk.Label(self, text="Must be a positive integer")
        self.low_warning = tk.Label(self, text="Must be a real number")
        self.high_warning = tk.Label(self, text="Must be a real number")
        self.low_high_warning = tk.Label(self, text="Min value must be smaller than max value")

        # Grid settings for the above warnings
        self.name_warning_gs = {"row":3, "column":2, "sticky":'EW'}
        self.ix_min_warning_gs = {"row":14, "column":2, "sticky":'EW'}
        self.length_warning_gs = {"row":15, "column":2, "sticky":'EW'}
        self.low_warning_gs = {"row":9, "column":3, "sticky":'EW'}
        self.high_warning_gs = {"row":10, "column":3, "sticky":'EW'}
        self.low_high_warning_gs = {"row":9, "column":3, "rowspan":2, "sticky":'EW'}

        # Responses to edits
        self.v1_array_wids[0].tk_wid.bind('<FocusOut>', self.name_handler)
        self.v1_array_wids[11].tk_wid.bind('<FocusOut>', self.ix_min_handler)
        self.v1_array_wids[12].tk_wid.bind('<FocusOut>', self.length_handler)
        self.v1_array_wids[6].tk_wid.bind('<FocusOut>', self.low_high_handler)
        self.v1_array_wids[7].tk_wid.bind('<FocusOut>', self.low_high_handler)

        # "Fill" to individual vars
        def var_fill_callback(ix):
            '''Helper function for buttons below'''
            return lambda : self.var_fill(ix)
        for ix in range(1, 11):
            tk.Button(self, text="Fill to vars", command=var_fill_callback(ix)).grid(row=ix+3, column=2)

        ttk.Separator(self, orient='horizontal').grid(row=i+1, column=0, columnspan=3, sticky='EW')
        i = i+1

        # Element browsers
        tk.Label(self, text="Choose elements:").grid(row=i+1, column=0, sticky='W')
        self.ele_name_button = tk.Button(self, text="Browse...",
                command=lambda : self.lat_browser())
        self.ele_name_button.grid(row=i+1, column=1, sticky='EW')

        ttk.Separator(self, orient='horizontal').grid(row=i+4, column=0, columnspan=3, sticky='EW')
        i = i+1

        # Individual variables
        self.var_dict = {}
        self.var_frame = tk.Frame(self)
        tk.Label(self, text="Variable:").grid(row=i+4, column=0, sticky='W')
        self.var_ix = tk.StringVar()
        self.ix_min = -1
        self.ix_max = -1
        self.var_chooser = ttk.Combobox(self, textvariable=self.var_ix,
                values=['PLEASE SPECIFY MIN/MAX INDICES'], state='readonly')
        self.var_chooser.bind('<<ComboboxSelected>>', self.make_var_frame)
        self.var_chooser.grid(row=i+4, column=1, sticky='EW')

        # Delete button
        tk.Button(self, text="DELETE THIS V1_ARRAY", fg='red', command=self.delete).grid(
                row=0, column=0, columnspan=3, sticky='EW')

        # Duplicate button
        tk.Button(self, text="Duplicate this v1_array", command=self.duplicate).grid(
                row=1, column=0, columnspan=3, sticky='EW')

        # Clone existing v1_array
        tk.Label(self, text="Clone existing v1:").grid(row=2, column=0, sticky='E')
        v1_list = self.pipe.cmd_in('python var_general').splitlines()
        for x in range(len(v1_list)):
            v1_list[x] = v1_list[x].split(';')[0]
        self.v1_clone = tk.StringVar()
        self.v1_clone.set(v1_list[0])
        tk.OptionMenu(self, self.v1_clone, *v1_list).grid(row=2, column=1, sticky='EW')
        tk.Button(self, text="Clone", command= lambda: self.clone(self.v1_clone.get())).grid(row=2, column=2, sticky='W')

        # Focus the v1 name widget
        self.v1_array_wids[0].tk_wid.focus_set()

    def delete(self, ask=True, event=None):
        '''
        Deletes this v1_array frame
        '''
        # Ask for confirmation
        ans = messagebox.askokcancel("Delete " + self.name, "Delete this v1_array and its associated data?", parent=self.parent)
        if not ans:
            return

        # Remove from parent
        ix = self.parent.v1_index
        if ix > 0:
            # Prefer to move to the tab to the left
            self.parent.notebook.select(ix-1)
        self.parent.v1_frame_list.pop(ix)

        # Destroy self
        self.destroy()

    def duplicate(self, event=None):
        '''
        Adds a new v1_frame to self.parent.v1_frame_list that is a copy of
        this frame, and changes focus to that frame
        '''
        # Don't run any handlers for this d1_frame
        self.handler_block = True
        self.parent.v1_frame_list.append(new_v1_frame(self.parent))
        ix = len(self.parent.v1_frame_list) - 1
        #self.d2_array.d1_index = ix
        # Copy properties into new frame
        self.parent.v1_frame_list[-1].name = self.name + '_copy'
        self.parent.v1_frame_list[-1].ix_min = self.ix_min
        self.parent.v1_frame_list[-1].ix_max = self.ix_max
        self.parent.v1_frame_list[-1].var_dict = self.var_dict
        for i in range(len(self.v1_array_wids)):
            if i == 0:
                self.parent.v1_frame_list[-1].v1_array_wids[i].tk_var.set(self.v1_array_wids[i].tk_var.get() + '_copy')
            else:
                self.parent.v1_frame_list[-1].v1_array_wids[i].tk_var.set(self.v1_array_wids[i].tk_var.get())
        # Run all input validation handlers
        self.parent.notebook.insert(ix, self.parent.v1_frame_list[-1])
        self.parent.notebook.select(ix)
        self.parent.tab_handler()
        self.update_idletasks()
        self.parent.v1_frame_list[-1].name_handler()
        self.parent.v1_frame_list[-1].ix_min_handler()
        self.parent.v1_frame_list[-1].length_handler()

    def make_var_frame(self, event=None):
        '''
        Adds an entry to self.var_dict if necessary, and updates
        self.var_frame to show the values for the current variable
        '''
        ix = self.var_ix.get()
        if ix == 'PLEASE SPECIFY MIN/MAX INDICES':
            return
        ix = int(ix)
        if ix not in self.var_dict.keys():
            self.var_dict[ix] = {}
        self.var_frame.destroy()
        self.var_frame = tk.Frame(self)
        def var_ttp(x):
            ''' Shortcut for commonly used construct '''
            return tk_tao_parameter(x, self.var_frame, self.pipe)
        # Parameters
        param_list = ['ele_name;STR;T;', 'universes;STR;T;', 'attribute;STR;T;',
                'weight;REAL;T;', 'step;REAL;T;', 'var^merit_type;ENUM;T;',
                'low_lim;REAL;T;', 'high_lim;REAL;T;', 'good_user;LOGIC;T;T',
                'key_bound;LOGIC;T;', 'key_delta;REAL;T;']
        param_list = list(map(str_to_tao_param, param_list))
        # Fill in defaults set at v1 level
        for i in range(1, len(param_list)):
            param_list[i].value = self.v1_array_wids[i].tk_var.get()
        # Set parameter values if specified
        for i in range(len(param_list)):
            if param_list[i].name in self.var_dict[ix].keys():
                param_list[i].value = self.var_dict[ix][param_list[i].name]
        # Create widgets
        self.var_wid_list = list(map(var_ttp, param_list))
        def var_writer(i):
            ''' Writes the contents of self.var_wid_list[i] into self.data_dict '''
            if i in [6,7]:
                if low_high_handler(grid=False):
                    return # Don't write to low/high limit if contents are invalid
            if self.var_wid_list[i].param.type == 'LOGIC':
                val = bool(self.var_wid_list[i].tk_var.get())
            else:
                val = self.var_wid_list[i].tk_var.get()
            self.var_dict[ix][self.var_wid_list[i].param.name] = val
        def var_writer_callback(i):
            ''' Callback for data_writer() '''
            return lambda *args : var_writer(i)

        # low/high validation
        def low_high_handler(self, grid=True, event=None):
            '''
            Verifies that the min and max values are set to good values
            returns 1 if they are not
            if grid is set False, messages will not be gridded/ungridded
            '''
            # Low limit
            fail = False
            if self.v1_array_wids[6].tk_var.get() != "":
                try:
                    low_lim = float(self.v1_array_wids[6].tk_var.get())
                except ValueError:
                    low_lim = None
                if low_lim == None:
                    self.low_high_warning_1.grid(row=7, column=2, sticky='W')
                    fail = True
                else:
                    self.low_high_warning_1.grid_forget()
            else:
                low_lim = None
            # High limit
            if self.v1_array_wids[7].tk_var.get() != "":
                try:
                    high_lim = float(self.v1_array_wids[7].tk_var.get())
                except ValueError:
                    high_lim = None
                if high_lim == None:
                    self.low_high_warning_2.grid(row=8, column=2, sticky='W')
                    fail = True
                else:
                    self.low_high_warning_2.grid_forget()
            else:
                high_lim = None
            # Make sure low < high
            if (low_lim != None) & (high_lim != None):
                if low_lim > high_lim:
                    self.low_high_warning_3.grid(row=7, column=2, sticky='W')
                    fail = True
                else:
                    self.low_high_warning_3.grid_forget()
            else:
                self.low_high_warning_3.grid_forget()

            if fail:
                return 1

        # Grid widgets
        low_high_warning_1 = tk.Label(self.var_frame, text="Must be a real number")
        low_high_warning_2 = tk.Label(self.var_frame, text="Must be a real number")
        low_high_warning_3 = tk.Label(self.var_frame, text="Low limit must be less than high limit")
        for i in range(len(self.var_wid_list)):
            self.var_wid_list[i].tk_label.grid(row=i, column=0, sticky='E')
            self.var_wid_list[i].tk_wid.grid(row=i, column=1, sticky='EW')
            # Write changes into self.var_dict
            self.var_wid_list[i].tk_var.trace('w', var_writer_callback(i))
        self.var_wid_list[6].tk_wid.bind('<FocusOut>', low_high_handler)
        self.var_wid_list[7].tk_wid.bind('<FocusOut>', low_high_handler)
        self.var_frame.grid(row=50, column=0, columnspan=3, sticky='EW')

    def var_fill(self, ix):
        '''
        "Fills" the value in self.v1_array_wids[ix] to all vars in self.var_dict
        Actually just erases the values in self.var_dict, which produces the same effect
        '''
        for i in self.var_dict.keys():
            param = self.v1_array_wids[ix].param.name
            if param in self.var_dict[i].keys():
                self.var_dict[i].pop(param)
            elif 'var^'+param in self.var_dict[i].keys():
                self.var_dict[i].pop('var^'+param)
        # Redraw datum frame to reflect changes
        try:
            int(self.var_ix.get())
            self.make_var_frame()
        except ValueError:
            pass

    def lat_browser(self):
        '''
        Opens a modified lattice table window to select elements en masse
        which should be either 'name', 'start_name', or 'ref_name'
        When the elements are chosen, ele_which will be set for each of this
        v1_array's variables sequentially
        '''
        # Make sure ix_min and length are not invalidly set
        if (self.ix_min_handler() == False) or (self.length_handler() == False):
            return
        # Autosize the array if self.length == -1
        if self.length == -1:
            autosize = True
        else:
            autosize = False
        win = tao_ele_browser(self.parent.root, self.pipe, self.name, self, 'var', autosize=autosize)

    def name_handler(self, event=None, strict=False):
        '''
        Checks whether the input name is valid, and
        sets the tab name to it if it is
        Returns True if the name is valid,
        False if the name is already in use by another v1,
        and None if the name is empty
        If strict is set True, a warning will be displayed
        for an empty name
        '''
        if self.handler_block:
            return
        name = self.v1_array_wids[0].tk_var.get().strip()
        if name == "":
            self.name_warning_invalid.grid_forget()
            self.name_warning_empty.grid_forget()
            # Warning in strict mode
            if strict:
                self.name_warning_empty.grid(**self.name_warning_gs)
            self.name = self.BLANK_TITLE
            self.parent.notebook.tab(self.parent.v1_index, text=self.BLANK_TITLE)
            return None
        # Name is nonempty
        # -> check if it's already in use
        i = 0
        for v1 in self.parent.v1_frame_list:
            if (v1.name == name) & (self != self.parent.v1_frame_list[i]):
                self.name_warning_empty.grid_forget()
                self.name_warning_invalid.grid(**self.name_warning_gs)
                return False
            i = i+1

        # Name is nonempty and valid
        self.parent.notebook.tab(self.parent.v1_index, text=name)
        self.name_warning_empty.grid_forget()
        self.name_warning_invalid.grid_forget()
        self.name = name
        if self.name == self.clone_of:
            return True # prevent repeatedly calling this method
        # Check if v1_array already exists in tao
        v1_list = self.pipe.cmd_in('python var_general').splitlines()
        for line in v1_list:
            if line.split(';')[0] == name:
                ans_var = tk.StringVar()
                tao_message_box(self.parent.root, self.parent, ans_var,
                        title=name+ " already exists",
                        message=name + " is already defined.    Would you like to edit its existing properties or overwrite them?",
                        choices=['Edit', 'Overwrite'])
                if ans_var.get() == 'Edit':
                    self.clone(self.name)
        return True

    def ix_min_handler(self, event=None, strict=False):
        '''
        Checks if the starting index is valid,
        and updates self.ix_min if it is.
        Returns True if a good index is set,
        False if a bad index is set
        (non-integer or greater than ix_max),
        and None if the ix_min field is empty
        If strict is set true, prints a warning message
        for an empty ix_min
        '''
        if self.handler_block:
            return
        # Check if min index is empty
        ix_min = self.v1_array_wids[11].tk_var.get().strip()
        if ix_min == "":
            self.ix_min = -1
            self.ix_min_warning.grid_forget()
            if strict:
                self.ix_min_warning.grid(**self.ix_min_warning_gs)
            #print("ix_min empty")
            return None
        # Check if min index is a non-negative integer
        try:
            ix_min = int(ix_min)
        except ValueError:
            ix_min = -1
        if ix_min < 0:
            self.ix_min_warning.grid(**self.ix_min_warning_gs)
            self.ix_min = -1
            #print("ix_min invalid")
            return False
        # ix_min is a good value -> update self.ix_min
        #print("valid ix_min")
        self.ix_min = ix_min
        self.ix_min_warning.grid_forget()
        # Update the data index range if self.length is set
        if self.length != -1:
            #print("updating ix_max and data_chooser")
            self.ix_max = self.ix_min + self.length - 1
            self.var_chooser.configure(values=list(range(self.ix_min, self.ix_max+1)))
        return True

    def length_handler(self, event=None, strict=False):
        '''
        Checks that the v1 array length has been set
        to a valid value, and updates self.length if it has.
        Returns True if the length has been set to a valid
        (positive integer) value, False if set to an invalid
        value, and None if left blank.
        '''
        if self.handler_block:
            #print("handler blocked")
            return
        # Check if length is empty
        length = self.v1_array_wids[12].tk_var.get().strip()
        if length == "":
            self.length = -1
            self.length_warning.grid_forget()
            if strict:
                self.length_warning.grid(**self.length_warning_gs)
            #print("length empty")
            return None
        # Check if length is a positive integer
        try:
            length = int(length)
        except ValueError:
            length = -1
        if length < 1:
            self.length_warning.grid(**self.length_warning_gs)
            self.length = -1
            #print("length invalid")
            return False
        # length is a good value -> update self.length
        self.length = length
        self.length_warning.grid_forget()
        #print("length valid")
        #print("self.ix_min:" + str(self.ix_min))
        # Update the data index range if self.ix_min is set
        if self.ix_min != -1:
            #print("updating self.ix_max and data_chooser")
            self.ix_max = self.ix_min + self.length - 1
            self.var_chooser.configure(values=list(range(self.ix_min, self.ix_max+1)))
        return True

    def low_high_handler(self, event=None, strict=False):
        '''
        Verifies that the min and max values are set to good values
        Returns True if they are,
        False if at least one is set invalidly,
        or None otherwise (at least one is blank)
        Prints an error for an empty low or high limit
        if strict is set True
        '''
        if self.handler_block:
            return
        self.low_warning.grid_forget()
        self.high_warning.grid_forget()
        self.low_high_warning.grid_forget()
        # Check if both limits are empty
        low_lim = self.v1_array_wids[6].tk_var.get()
        high_lim = self.v1_array_wids[7].tk_var.get()
        if (low_lim == "") and (high_lim == ""): #both are empty
            if strict:
                self.low_warning.grid(**self.low_warning_gs)
                self.high_warning.grid(**self.high_warning_gs)
            return None
        elif (low_lim == ""): #low is empty
            try:
                high_lim = float(high_lim)
            except ValueError:
                high_lim = None
            if strict:
                self.low_warning.grid(**self.low_warning_gs)
            if high_lim == None:
                self.high_warning.grid(**self.high_warning_gs)
                return False
            return None
        elif (high_lim == ""): #high is empty
            try:
                low_lim = float(low_lim)
            except ValueError:
                low_lim = None
            if strict:
                self.high_warning.grid(**self.high_warning_gs)
            if low_lim == None:
                self.low_warning.grid(**self.low_warning_gs)
                return False
            return None
        else: #neither are empty
            try:
                low_lim = float(low_lim)
            except ValueError:
                low_lim = None
                self.low_warning.grid(**self.low_warning_gs)
            try:
                high_lim = float(high_lim)
            except ValueError:
                high_lim = None
                self.high_warning.grid(**self.high_warning_gs)
            if None in [low_lim, high_lim]:
                return False
            # Check that low_lim < high_lim
            if not (low_lim < high_lim):
                self.low_high_warning.grid(**self.low_high_warning_gs)
                return False
            return True

    def clone(self, v1_array):
        '''
        Clones the existing v1_array into this v1_frame
        '''
        # Read existing var info from tao
        var_general = self.pipe.cmd_in('python var_general').splitlines()
        for line in var_general:
            if line.find(v1_array) == 0:
                break
        line = line.split(';')
        # set name
        self.v1_array_wids[0].tk_var.set(v1_array)
        self.clone_of = v1_array
        self.name_handler()
        # set min/max indices
        self.v1_array_wids[11].tk_var.set(line[2])
        self.v1_array_wids[12].tk_var.set(line[3])
        self.ix_min_handler()
        self.length_handler()
        # copy inidividual variable info
        var_params = ['ele_name', 'attribute', 'weight', 'step', 'var^merit_type', 'low_lim', 'high_lim', 'good_user', 'key_bound', 'key_delta']
        self.var_dict = {}
        for i in range(self.ix_min, self.ix_max+1):
            self.var_dict[i] = {}
            var_info = self.pipe.cmd_in('python var ' + v1_array + '[' + str(i) + ']')
            var_info = var_info.splitlines()
            for p in var_params:
                found=False
                for line in var_info:
                    if line.find(p) == 0:
                        found=True
                        break
                if found:
                    self.var_dict[i][p] = line.split(';')[3].strip()
            # Special case: attribute (vs attrib_name)
            found=False
            for line in var_info:
                if line.find('attrib_name') == 0:
                    found=True
                    break
            if found:
                self.var_dict[i]['attribute'] = line.split(';')[3].strip()
            # Special case: universes
            unis = []
            for line in self.pipe.cmd_in('python var ' + v1_array + '[' + str(i) + '] slaves').splitlines():
                u = int(line.split(';')[0])
                if u not in unis:
                    unis.append(u)
            unis.sort()
            universes = ""
            for u in unis:
                universes += str(u) + ','
            universes = universes[:-1] # remove extra comma
            self.var_dict[i]['universes'] = universes
        # Set v1 defaults to var_dict[self.ix_min] values
        for i in range(1, len(var_params)):
            self.v1_array_wids[i+1].tk_var.set(self.var_dict[self.ix_min][var_params[i]])




'''
This package implements a GUI for tao using tkinter, including interactive
plots with matplotlib.  Matplotlib is required for the GUI.
The GUI is organized into one root window from which various other
toplevel windows can be spawned, including plots, element windows, and tables
for data, variables, and lattice elements.  The GUI also supports defining
new data, variables, and plot templates on the fly.
To start the gui, just run

python -m pytao.gui <args>

on the command line.  <args> can be any switches that tao supports.
Additionally, the GUI supports the -gui_init switch, which takes a
gui.init file as its argument (see GUI documentation for more info)
'''
from .main import tao_root_window



import pytest
import os


@pytest.fixture
def rootdir():
    return os.path.dirname(os.path.abspath(__file__))


@pytest.fixture
def config_file(rootdir):
    return open(f"{rootdir}/test_files/iris_config.yml", "r")




import os
from pytao import Tao
from pytao import interface_commands


def test_beam_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/csr_beam_tracking/tao.init -noplot'))
    ret = interface_commands.beam(tao, ix_uni='1')
            
        
def test_beam_init_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/csr_beam_tracking/tao.init -noplot'))
    ret = interface_commands.beam_init(tao, ix_uni='1')
            
        
def test_bmad_com_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.bmad_com(tao)
            
        
def test_branch1_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.branch1(tao, ix_uni='1', ix_branch='0')
            
        
def test_bunch_params_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/csr_beam_tracking/tao.init -noplot'))
    ret = interface_commands.bunch_params(tao, ele_id='end', which='model')
            
        
def test_bunch1_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/csr_beam_tracking/tao.init -noplot'))
    ret = interface_commands.bunch1(tao, ele_id='end', coordinate='x', which='model', ix_bunch='1')
            
        
def test_building_wall_list_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_wall -noplot'))
    ret = interface_commands.building_wall_list(tao, ix_section='')
            
        
def test_building_wall_list_2():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_wall -noplot'))
    ret = interface_commands.building_wall_list(tao, ix_section='1')
            
        
def test_building_wall_graph_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_wall -noplot'))
    ret = interface_commands.building_wall_graph(tao, graph='floor_plan.g')
            
        
def test_building_wall_point_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_wall -noplot'))
    ret = interface_commands.building_wall_point(tao, ix_section='1', ix_point='1', z='0', x='0', radius='0', z_center='0', x_center='0')
            
        
def test_building_wall_section_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.building_wall_section(tao, ix_section='1', sec_name='test', sec_constraint='none')
            
        
def test_constraints_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_optics_matching -noplot'))
    ret = interface_commands.constraints(tao, who='data')
            
        
def test_constraints_2():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.constraints(tao, who='var')
            
        
def test_data_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_optics_matching -noplot'))
    ret = interface_commands.data(tao, ix_uni='', d2_name='twiss', d1_name='end', dat_index='1')
            
        
def test_data_2():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_optics_matching -noplot'))
    ret = interface_commands.data(tao, ix_uni='1', d2_name='twiss', d1_name='end', dat_index='1')
            
        
def test_data_d2_create_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_optics_matching -noplot'))
    ret = interface_commands.data_d2_create(tao, ix_uni='1', d2_name='orbit', n_d1_data='2', d_data_arrays_name_min_max='x^^0^^45^^y^^1^^47')
            
        
def test_data_d2_destroy_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.data_d2_destroy(tao, d2_name='orbit')
            
        
def test_data_d2_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_optics_matching -noplot'))
    ret = interface_commands.data_d2(tao, ix_uni='1', d2_name='twiss')
            
        
def test_data_d_array_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_optics_matching -noplot'))
    ret = interface_commands.data_d_array(tao, ix_uni='1', d2_name='twiss', d1_name='end')
            
        
def test_data_d1_array_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_optics_matching -noplot'))
    ret = interface_commands.data_d1_array(tao, ix_uni='1', d2_datum='twiss')
            
        
def test_data_parameter_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_optics_matching -noplot'))
    ret = interface_commands.data_parameter(tao, data_array='twiss.end', parameter='model_value')
            
        
def test_data_d2_array_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.data_d2_array(tao, ix_uni='1')
            
        
def test_data_set_design_value_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_optics_matching -noplot'))
    ret = interface_commands.data_set_design_value(tao)
            
        
def test_datum_create_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_optics_matching -noplot'))
    ret = interface_commands.datum_create(tao, datum_name='twiss.end[6]', data_type='beta.y', ele_ref_name='', ele_start_name='', ele_name='P1', merit_type='target', meas='0', good_meas='T', ref='0', good_ref='T', weight='0.3', good_user='T', data_source='lat', eval_point='END', s_offset='0', ix_bunch='1', invalid_value='0')
            
        
def test_datum_has_ele_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_optics_matching -noplot'))
    ret = interface_commands.datum_has_ele(tao, datum_type='twiss.end')
            
        
def test_derivative_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_optics_matching -noplot'))
    ret = interface_commands.derivative(tao)
            
        
def test_ele_ac_kicker_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.ele_ac_kicker(tao, ele_id='1@0>>1', which='model')
            
        
def test_ele_cartesian_map_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_em_field -noplot'))
    ret = interface_commands.ele_cartesian_map(tao, ele_id='1@0>>1', which='model', index='1', who='base')
            
        
def test_ele_chamber_wall_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_wall3d -noplot'))
    ret = interface_commands.ele_chamber_wall(tao, ele_id='1@0>>1', which='model', index='1', who='x')
            
        
def test_ele_control_var_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.ele_control_var(tao, ele_id='1@0>>873', which='model')
            
        
def test_ele_cylindrical_map_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_em_field -noplot'))
    ret = interface_commands.ele_cylindrical_map(tao, ele_id='1@0>>5', which='model', index='1', who='base')
            
        
def test_ele_elec_multipoles_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.ele_elec_multipoles(tao, ele_id='1@0>>1', which='model')
            
        
def test_ele_floor_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.ele_floor(tao, ele_id='1@0>>1', which='model', where='')
            
        
def test_ele_floor_2():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.ele_floor(tao, ele_id='1@0>>1', which='model', where='center')
            
        
def test_ele_grid_field_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_grid -noplot'))
    ret = interface_commands.ele_grid_field(tao, ele_id='1@0>>1', which='model', index='1', who='base')
            
        
def test_ele_gen_attribs_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.ele_gen_attribs(tao, ele_id='1@0>>1', which='model')
            
        
def test_ele_head_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.ele_head(tao, ele_id='1@0>>1', which='model')
            
        
def test_ele_lord_slave_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.ele_lord_slave(tao, ele_id='1@0>>1', which='model')
            
        
def test_ele_mat6_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.ele_mat6(tao, ele_id='1@0>>1', which='model', who='mat6')
            
        
def test_ele_methods_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.ele_methods(tao, ele_id='1@0>>1', which='model')
            
        
def test_ele_multipoles_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.ele_multipoles(tao, ele_id='1@0>>1', which='model')
            
        
def test_ele_orbit_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.ele_orbit(tao, ele_id='1@0>>1', which='model')
            
        
def test_ele_param_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_photon -noplot'))
    ret = interface_commands.ele_param(tao, ele_id='1@0>>1', which='model', who='orbit.vec.1')
            
        
def test_ele_photon_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_photon -noplot'))
    ret = interface_commands.ele_photon(tao, ele_id='1@0>>1', which='model', who='base')
            
        
def test_ele_spin_taylor_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_spin -noplot'))
    ret = interface_commands.ele_spin_taylor(tao, ele_id='1@0>>2', which='model')
            
        
def test_ele_taylor_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_taylor -noplot'))
    ret = interface_commands.ele_taylor(tao, ele_id='1@0>>34', which='model')
            
        
def test_ele_taylor_field_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_em_field -noplot'))
    ret = interface_commands.ele_taylor_field(tao, ele_id='1@0>>9', which='model', index='1', who='terms')
            
        
def test_ele_twiss_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.ele_twiss(tao, ele_id='1@0>>1', which='model')
            
        
def test_ele_wake_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_wake -noplot'))
    ret = interface_commands.ele_wake(tao, ele_id='1@0>>1', which='model', who='sr_long')
            
        
def test_ele_wall3d_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_wall3d -noplot'))
    ret = interface_commands.ele_wall3d(tao, ele_id='1@0>>1', which='model', index='1', who='table')
            
        
def test_evaluate_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.evaluate(tao, expression='data::cbar.11[1:10]|model')
            
        
def test_em_field_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.em_field(tao, ele_id='1@0>>22', which='model', x='0', y='0', z='0', t_or_z='0')
            
        
def test_enum_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.enum(tao, enum_name='tracking_method')
            
        
def test_floor_plan_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_optics_matching -noplot'))
    ret = interface_commands.floor_plan(tao, graph='r13.g')
            
        
def test_floor_orbit_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_floor_orbit -noplot'))
    ret = interface_commands.floor_orbit(tao, graph='r33.g')
            
        
def test_tao_global_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.tao_global(tao)
            
        
def test_help_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.help(tao)
            
        
def test_inum_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.inum(tao, who='ix_universe')
            
        
def test_lat_calc_done_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.lat_calc_done(tao, branch_name='1@0')
            
        
def test_lat_ele_list_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.lat_ele_list(tao, branch_name='1@0')
            
        
def test_lat_branch_list_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.lat_branch_list(tao, ix_uni='1')
            
        
def test_lat_list_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.lat_list(tao, ix_uni='1', ix_branch='0', elements='Q*', which='model', who='orbit.floor.x')
            
        
def test_lat_list_2():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.lat_list(tao, ix_uni='1', ix_branch='0', elements='Q*', which='design', who='ele.ix_ele')
            
        
def test_lat_param_units_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.lat_param_units(tao, param_name='L')
            
        
def test_matrix_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.matrix(tao, ele1_id='1@0>>q01w|design', ele2_id='q02w')
            
        
def test_merit_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.merit(tao)
            
        
def test_orbit_at_s_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.orbit_at_s(tao, ix_uni='1', ele='10', s_offset='0.7', which='model')
            
        
def test_place_buffer_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.place_buffer(tao)
            
        
def test_plot_curve_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_optics_matching -noplot'))
    ret = interface_commands.plot_curve(tao, curve_name='r13.g.a')
            
        
def test_plot_lat_layout_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.plot_lat_layout(tao, ix_uni='1', ix_branch='0')
            
        
def test_plot_list_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.plot_list(tao, r_or_g='r')
            
        
def test_plot_graph_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_optics_matching -noplot'))
    ret = interface_commands.plot_graph(tao, graph_name='beta.g')
            
        
def test_plot_histogram_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_optics_matching -noplot'))
    ret = interface_commands.plot_histogram(tao, curve_name='r33.g.x')
            
        
def test_plot_plot_manage_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_optics_matching -noplot'))
    ret = interface_commands.plot_plot_manage(tao, plot_location='@T1', plot_name='beta', n_graph='1', graph1_name='g1', graph2_name='g2', graphN_name='gN')
            
        
def test_plot_curve_manage_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_optics_matching -noplot'))
    ret = interface_commands.plot_curve_manage(tao, graph_name='beta.g', curve_index='1', curve_name='r13.g.a')
            
        
def test_plot_graph_manage_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_optics_matching -noplot'))
    ret = interface_commands.plot_graph_manage(tao, plot_name='beta', graph_index='1', graph_name='beta.g')
            
        
def test_plot_line_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_plot_line -external_plotting -noplot'))
    ret = interface_commands.plot_line(tao, region_name='beta', graph_name='g', curve_name='a', x_or_y='')
            
        
def test_plot_line_2():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_plot_line -external_plotting -noplot'))
    ret = interface_commands.plot_line(tao, region_name='beta', graph_name='g', curve_name='a', x_or_y='y')
            
        
def test_plot_symbol_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_plot_line -external_plotting -noplot'))
    ret = interface_commands.plot_symbol(tao, region_name='r13', graph_name='g', curve_name='a', x_or_y='')
            
        
def test_plot_symbol_2():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_plot_line -external_plotting -noplot'))
    ret = interface_commands.plot_symbol(tao, region_name='r13', graph_name='g', curve_name='a', x_or_y='y')
            
        
def test_plot_transfer_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_optics_matching -noplot'))
    ret = interface_commands.plot_transfer(tao, from_plot='r13', to_plot='r23')
            
        
def test_plot1_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_optics_matching -noplot'))
    ret = interface_commands.plot1(tao, name='beta')
            
        
def test_ptc_com_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.ptc_com(tao)
            
        
def test_ring_general_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.ring_general(tao, ix_uni='1', ix_branch='0', which='model')
            
        
def test_shape_list_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.shape_list(tao, who='floor_plan')
            
        
def test_shape_manage_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.shape_manage(tao, who='floor_plan', index='1', add_or_delete='add')
            
        
def test_shape_pattern_list_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_shape -noplot'))
    ret = interface_commands.shape_pattern_list(tao, ix_pattern='')
            
        
def test_shape_pattern_manage_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_shape -noplot'))
    ret = interface_commands.shape_pattern_manage(tao, ix_pattern='1', pat_name='new_pat', pat_line_width='1')
            
        
def test_shape_pattern_point_manage_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_shape -noplot'))
    ret = interface_commands.shape_pattern_point_manage(tao, ix_pattern='1', ix_point='1', s='0', x='0')
            
        
def test_shape_set_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.shape_set(tao, who='floor_plan', shape_index='1', ele_name='Q1', shape='circle', color='', shape_size='', type_label='', shape_draw='', multi_shape='', line_width='')
            
        
def test_show_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.show(tao, line='-python')
            
        
def test_species_to_int_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.species_to_int(tao, species_str='electron')
            
        
def test_species_to_str_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.species_to_str(tao, species_int='-1')
            
        
def test_spin_polarization_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.spin_polarization(tao, ix_uni='1', ix_branch='0', which='model')
            
        
def test_spin_resonance_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.spin_resonance(tao, ix_uni='1', ix_branch='0', which='model')
            
        
def test_super_universe_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.super_universe(tao)
            
        
def test_twiss_at_s_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.twiss_at_s(tao, ix_uni='1', ele='10', s_offset='0.7', which='model')
            
        
def test_universe_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.universe(tao, ix_uni='1')
            
        
def test_var_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_optics_matching -noplot'))
    ret = interface_commands.var(tao, var='quad[1]', slaves='')
            
        
def test_var_2():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_optics_matching -noplot'))
    ret = interface_commands.var(tao, var='quad[1]', slaves='slaves')
            
        
def test_var_create_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/tao.init_optics_matching -noplot'))
    ret = interface_commands.var_create(tao, var_name='quad[1]', ele_name='Q1', attribute='L', universes='1', weight='0.001', step='0.001', low_lim='-10', high_lim='10', merit_type='', good_user='T', key_bound='T', key_delta='0.01')
            
        
def test_var_general_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.var_general(tao)
            
        
def test_var_v_array_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.var_v_array(tao, v1_var='quad_k1')
            
        
def test_var_v1_array_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.var_v1_array(tao, v1_var='quad_k1')
            
        
def test_var_v1_create_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.var_v1_create(tao, v1_name='quad_k1', n_var_min='0', n_var_max='45')
            
        
def test_var_v1_destroy_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.var_v1_destroy(tao, v1_datum='quad_k1')
            
        
def test_wave_1():
    
    tao = Tao(os.path.expandvars('-init $ACC_ROOT_DIR/regression_tests/python_test/cesr/tao.init -noplot'))
    ret = interface_commands.wave(tao, what='params')
            
        


# Functions to handle CSR
import numpy as np

def parse_csr_wake(file, prepend_index=True):
    """
    Parses CSR wake file and returns a dict of elements with
    
    
    ix_ele: index of the element
    
    s_positions: np.array of s possitions correponding to the steps in data
    
    labels: list of string labels for columns
    
    data: np.array of (step, bin, column)
        column 0: Z
        column 1: Charge/Meter
        column 2: CSR_Kick/m
        column 3: I_CSR/m
        column 4: S_Source
    """
    with open(file) as f:
        eles = {}
        line = f.readline()

        data = None # Start 
        while line:
        
            
            # Check for new ele
            if line.startswith('!------------------------------------------------------------'):
                line = f.readline()    
                _, ix, ele_name = line.split()
                ix = int(ix)
                
                # Add index to key for uniqueness
                if prepend_index:
                    key = f'{ix}:{ele_name}'
                else:
                    key = ele_name
                
                eles[key] = {'data':[], 's_positions':[], 'ix_ele':ix,
                                  'ele_name':ele_name,
                                  'labels':['z',
                                            'charge_per_meter',
                                            'csr_kick_per_meter',
                                            'I_csr_per_meter',
                                            's_source']}
                # Current ele
                ele = eles[key]
                
                # point to current data
                ele_steps= ele['data'] = []
                
              #  print('new ele', ele_name)
                line = f.readline()
                continue
            #print(line)
            
            # Check for new step
            if line.startswith('!#-----------------------------'):
                
                line = f.readline()
                step = int(line.split('index:')[-1])
                line = f.readline()
                s_position = float(line.split()[-1])
                f.readline() # header
                
                # new data 
                data = []
                ele_steps.append(data)
                ele['s_positions'].append(s_position)
               # print('data: ', data)
                
               # print('new step:', step)
                line = f.readline()
                continue
                
           # print(line)    
            data.append(line)
                
            
            
            line = f.readline()
                        
    # Cast to numpy
    for _, ele in eles.items():
        ele['data'] = np.array([np.loadtxt(data) for data in ele['data']])
        ele['s_positions'] = np.array(ele['s_positions'] )
    return eles



def process_csr_wake_data(data):
    """
    Process csr_wake data by concatenating the data. 
    
    Returns a dict of np.array with labels and shapes:
    
        s_position          (steps,)      
        ix_ele              (steps,)
        z                   (steps, bins)
        charge_per_meter    (steps, bins)
        csr_kick_per_meter  (steps, bins)
        I_csr_per_meter     (steps, bins)
        s_source            (steps, bins)
    
    and a dict:
        ele_name_from_ix
    that returns the ele_name from an ix_ele for convenience. 
    
    
    """
    
    # Sort by ix_ele
    skeys = sorted(data, key=lambda n: data[n]['ix_ele'] )
    labels = data[skeys[0]]['labels'].tolist()
    #return labels

    output = {}
    # Concatenate all
    dat = np.concatenate([data[key]['data'] for key in skeys ])
    output['s_position'] = np.concatenate([data[key]['s_positions'] for key in skeys ])
    
    # Fill out ix_ele
    output['ix_ele'] = np.concatenate([np.full(len(data[key]['s_positions']),  data[key]['ix_ele']) for key in skeys])
    
    # ix-> name mapping
    output['ele_name_from_ix'] = {data[key]['ix_ele']:key for key in data }
    
    # Split columns
    for i, key in enumerate(labels):
        output[key] = dat[:,:,i]
    
    return output
    

    
def csr_wake_stats_at_step(data, step=0):
    """
    Calculates stats from a single step of processed csr_wake data.
    
    Useful for plotting.
    
    Returns tuple:
    
        avkick, stdkick, avz, stdz
    
    """
 
    z = data['z'][step,:]

    
    zmax = z.max()
    zmin = z.min()
    nz = len(z)
    dz = z.ptp()/(nz-1)
    density = data['charge_per_meter'][step,:]
    dsum = np.sum(density)
    qtot = dsum*dz
    density = density/dsum
 
    
    avz = np.sum(z*density)
    stdz = np.sqrt(np.sum(z**2*density) - avz**2)
    
    kick = data['csr_kick_per_meter'][step,:]
    avkick = np.sum(kick*density)
    stdkick = np.sqrt(np.sum( kick**2*density) - avkick**2)
    
    return avkick, stdkick, avz, stdz    


def write_csr_wake_data_h5(h5, data, name=None):
    """
    Write parsed csr_wake data to an open h5 handle
    """
    if name:
        h5 = h5.create_group(name)
    
    for ele in data:
        g = h5.create_group(ele)
        for k, v in data[ele].items():
            # Write np.array as datasets, otherwise to attributes
            if isinstance(v, np.ndarray):
                g[k] = v
            else:
                g.attrs[k] = v
                
def read_csr_wake_data_h5(h5, name=None):
    """
    Read csr_wake data from h5 file.
    
    See: write_csr_wake_data_h5
    """
    
    
    if name:
        h5 = h5[name]
    dat = {}
    for g in h5:
        dat[g] = {}
        ele = dat[g]
        # Read all datasets and attributes
        for key in h5[g]:
            ele[key] = h5[g][key][:]
        ele.update(dict(h5[g].attrs))
    return dat






from .core import run_tao
from ..misc.csr import parse_csr_wake, write_csr_wake_data_h5

from .tools import fingerprint

from h5py import File

import os


def evaluate_tao(settings,
                     run_commands=['set global track_type=beam'],
                     expressions=['lat::orbit.x[end]'],
                     input_file=None,
                     ploton=False,
                    
                     beam_archive_path=None,
                     archive_csr_wake=False,
                     workdir=None,
                     so_lib='',
                     verbose=False):
    """
    
    settings: dict of set_command:value where set_command is a string.
        Example:
            'global:track_type':'beam'
                will issue command:
            set global track_type = beam
    
    run_commands: list of command strings that will be executed.
    
    expressions: list of expression strings that will be used to form the output.
    
    beam_archive_path: if given, the all of the saved beams will be written to
        a file named by a fingerprint (hash) of the inputs into path beam_archive_path.
        This uses the command:
            write beam -at *
        which writes ALL of the bunches that are saved using the beam_saved_at list in beam_init.
        
    archive_csr_wake: if given, will look for csr_wake.dat, parse, and archive to the h5 file above.
    
    Returns a dict of expression:value, according to the expressions above, as well as 
        beam_archive if a  beam_archive_path was given.
    
    
    
    Example:
    
    evaluate_tao(settings={}, 
                   input_file=tao.init', 
                   run_commands=['set global track_type=beam'],
                   expressions = ['lat::orbit.x[FF.PIP02A]', 'beam::norm_emit.x[end]'],
                   ploton=False, 
                   beam_archive_path = '.')
                   
    Returns:
    
        {'lat::orbit.x[FF.PIP02A]': 0.0,
         'beam::norm_emit.x[end]': 9.9982321550206e-07,
         'beam_archive': /path/to/bmad_beam_7fd6d30ac45a3d8c0d45112f4b569dee.h5'}
    
    
    
    See: run_tao
    
    """

    M = run_tao(settings=settings,
                run_commands=run_commands,
                input_file=input_file,
                ploton=ploton,
                workdir=workdir,
                so_lib=so_lib,
                verbose=verbose)
    
    
    output = {}
    
    for expression in expressions:
        try:
            val = M.evaluate(expression)
        except:
            print(f'error with {expression}')
            val = None
        output[expression] = val    
    
    if beam_archive_path:
        ff = fingerprint({'input_file':input_file, 'settings':settings})
        beam_archive_path = os.path.expandvars(beam_archive_path)
        beam_archive = os.path.abspath(os.path.join(beam_archive_path, f'bmad_beam_{ff}'+'.h5'))
        if verbose:
            print('Archiving beam to', beam_archive)
        M.cmd(f'write beam -at * {beam_archive}')    
        output['beam_archive'] = beam_archive
        

        # Reopen and attach settings
        assert os.path.exists(beam_archive), 'No archive was written. Perhaps there was no beam?'
        
        with File(beam_archive, 'r+') as h5:
            # Input
            g = h5.create_group('input')
            g.attrs['input_file'] = input_file
            
            #g.attrs['input_file'] = input_file
            
            
            # Settings
            g = h5.create_group('settings')
            for k, v in settings.items():
                g.attrs[k] = v
        
            g = h5.create_group('expressions')
            for k, v in output.items():
                if v:
                    g.attrs[k] = v
                    
            # CSR wake
            csr_wake_file = os.path.join(M.path, 'csr_wake.dat')

            if archive_csr_wake and os.path.exists(csr_wake_file):
                csr_wake_data = parse_csr_wake(csr_wake_file)
                write_csr_wake_data_h5(h5, csr_wake_data, name='csr_wake')
                

        
    
    return output


